"""
Functions for network matched-filter detection of seismic data.

Designed to cross-correlate templates generated by template_gen function
with data and output the detections.

:copyright:
    EQcorrscan developers.

:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import copy
import getpass
import glob
import os
import pickle
import shutil
import tarfile
import tempfile
import traceback
import uuid
import logging

from multiprocessing import Process, Queue, cpu_count
from queue import Empty

from obspy import Catalog, Stream, read, read_events
from obspy.core.event import Comment, CreationInfo
from obspy.clients.fdsn import Client as FDSNClient

from eqcorrscan.core import template_gen
from eqcorrscan.core.match_filter.template import (
    Template, quick_group_templates)
from eqcorrscan.core.match_filter.party import Party
from eqcorrscan.core.match_filter.family import Family
from eqcorrscan.core.match_filter.helpers import (
    _safemembers, _par_read, get_waveform_client,
    _remove_duplicates, _moveout)
from eqcorrscan.core.match_filter.helpers.tribe import _wildcard_fill

from eqcorrscan.utils.pre_processing import (
    _quick_copy_stream, _prep_data_for_correlation)


Logger = logging.getLogger(__name__)


class Tribe(object):
    """
    Holder for multiple templates.

    :type templates: List of Template
    :param templates: The templates within the Tribe.
    """
    _timeout = 1

    def __init__(self, templates=None):
        self.templates = []
        if isinstance(templates, Template):
            templates = [templates]
        if templates:
            self.templates.extend(templates)
        # Managers for Processes and Queues to be killed on errors
        self._processes = dict()
        self._queues = dict()
        # Assign unique ids
        self.__unique_ids()

    def __repr__(self):
        """
        Print information about the tribe.

        .. rubric:: Example

        >>> tribe = Tribe(templates=[Template(name='a')])
        >>> print(tribe)
        Tribe of 1 templates
        """
        return 'Tribe of %i templates' % self.__len__()

    def __add__(self, other):
        """
        Add two Tribes or a Tribe and a Template together. '+'

        .. rubric:: Example

        >>> tribe = Tribe(templates=[Template(name='a')])
        >>> tribe_ab = tribe + Tribe(templates=[Template(name='b')])
        >>> print(tribe_ab)
        Tribe of 2 templates
        >>> tribe_abc = tribe_ab + Template(name='c')
        >>> print(tribe_abc)
        Tribe of 3 templates
        """
        return self.copy().__iadd__(other)

    def __iadd__(self, other):
        """
        Add in place: '+='

        .. rubric:: Example

        >>> tribe = Tribe(templates=[Template(name='a')])
        >>> tribe += Tribe(templates=[Template(name='b')])
        >>> print(tribe)
        Tribe of 2 templates
        >>> tribe += Template(name='c')
        >>> print(tribe)
        Tribe of 3 templates
        """
        assert isinstance(other, (Tribe, Template)), \
            "Must be either Template or Tribe"
        self.__unique_ids(other)
        if isinstance(other, Tribe):
            self.templates += other.templates
        elif isinstance(other, Template):
            self.templates.append(other)
        return self

    def __eq__(self, other):
        """
        Test for equality. Rich comparison operator '=='

        .. rubric:: Example

        >>> tribe_a = Tribe(templates=[Template(name='a')])
        >>> tribe_b = Tribe(templates=[Template(name='b')])
        >>> tribe_a == tribe_b
        False
        >>> tribe_a == tribe_a
        True
        """
        if self.sort().templates != other.sort().templates:
            return False
        return True

    def __ne__(self, other):
        """
        Test for inequality. Rich comparison operator '!='

        .. rubric:: Example

        >>> tribe_a = Tribe(templates=[Template(name='a')])
        >>> tribe_b = Tribe(templates=[Template(name='b')])
        >>> tribe_a != tribe_b
        True
        >>> tribe_a != tribe_a
        False
        """
        return not self.__eq__(other)

    def __len__(self):
        """
        Number of Templates in Tribe. len(tribe)

        .. rubric:: Example

        >>> tribe_a = Tribe(templates=[Template(name='a')])
        >>> len(tribe_a)
        1
        """
        return len(self.templates)

    def __iter__(self):
        """
        Iterator for the Tribe.
        """
        return list(self.templates).__iter__()

    def __getitem__(self, index):
        """
        Support slicing to get Templates from Tribe.

        .. rubric:: Example

        >>> tribe = Tribe(templates=[Template(name='a'), Template(name='b'),
        ...                          Template(name='c')])
        >>> tribe[1] # doctest: +NORMALIZE_WHITESPACE
        Template b:
         0 channels;
         lowcut: None Hz;
         highcut: None Hz;
         sampling rate None Hz;
         filter order: None;
         process length: None s
        >>> tribe[0:2]
        Tribe of 2 templates
        >>> tribe["d"]
        []
        """
        if isinstance(index, slice):
            return self.__class__(templates=self.templates.__getitem__(index))
        elif isinstance(index, int):
            return self.templates.__getitem__(index)
        else:
            _index = [i for i, t in enumerate(self.templates)
                      if t.name == index]
            try:
                return self.templates.__getitem__(_index[0])
            except IndexError:
                Logger.warning('Template: %s not in tribe' % index)
                return []

    def __unique_ids(self, other=None):
        """ Check that template names are unique. """
        template_names = [t.name for t in self.templates]
        if other:
            assert isinstance(other, (Template, Tribe)), (
                "Can only test against tribes or templates")
            if isinstance(other, Template):
                template_names.append(other.name)
            else:
                template_names.extend([t.name for t in other])

        unique_names = set(template_names)
        if len(unique_names) < len(template_names):
            non_unique_names = [name for name in unique_names
                                if template_names.count(name) > 1]
            raise NotImplementedError(
                "Multiple templates found with the same name. Template names "
                "must be unique. Non-unique templates: "
                f"{', '.join(non_unique_names)}")
        return

    @property
    def _stream_dir(self):
        """ Location for temporary streams """
        return f".streams_{os.getpid()}"

    def sort(self):
        """
        Sort the tribe, sorts by template name.

        .. rubric:: Example

        >>> tribe = Tribe(templates=[Template(name='c'), Template(name='b'),
        ...                          Template(name='a')])
        >>> tribe.sort()
        Tribe of 3 templates
        >>> tribe[0] # doctest: +NORMALIZE_WHITESPACE
        Template a:
         0 channels;
         lowcut: None Hz;
         highcut: None Hz;
         sampling rate None Hz;
         filter order: None;
         process length: None s
        """
        self.templates = sorted(self.templates, key=lambda x: x.name)
        return self

    def select(self, template_name):
        """
        Select a particular template from the tribe.

        :type template_name: str
        :param template_name: Template name to look-up
        :return: Template

        .. rubric:: Example

        >>> tribe = Tribe(templates=[Template(name='c'), Template(name='b'),
        ...                          Template(name='a')])
        >>> tribe.select('b') # doctest: +NORMALIZE_WHITESPACE
        Template b:
         0 channels;
         lowcut: None Hz;
         highcut: None Hz;
         sampling rate None Hz;
         filter order: None;
         process length: None s
        """
        return [t for t in self.templates if t.name == template_name][0]

    def remove(self, template):
        """
        Remove a template from the tribe.

        :type template: :class:`eqcorrscan.core.match_filter.Template`
        :param template: Template to remove from tribe

        .. rubric:: Example

        >>> tribe = Tribe(templates=[Template(name='c'), Template(name='b'),
        ...                          Template(name='a')])
        >>> tribe.remove(tribe.templates[0])
        Tribe of 2 templates
        """
        self.templates = [t for t in self.templates if t != template]
        return self

    def copy(self):
        """
        Copy the Tribe.

        .. rubric:: Example

        >>> tribe_a = Tribe(templates=[Template(name='a')])
        >>> tribe_b = tribe_a.copy()
        >>> tribe_a == tribe_b
        True
        """
        # We can't copy processes, so we need to just copy the templates
        other = Tribe(copy.deepcopy(self.templates))
        return other

    def write(self, filename, compress=True, catalog_format="QUAKEML"):
        """
        Write the tribe to a file using tar archive formatting.

        :type filename: str
        :param filename:
            Filename to write to, if it exists it will be appended to.
        :type compress: bool
        :param compress:
            Whether to compress the tar archive or not, if False then will
            just be files in a folder.
        :type catalog_format: str
        :param catalog_format:
            What format to write the detection-catalog with. Only Nordic,
            SC3ML, QUAKEML are supported. Note that not all information is
            written for all formats (QUAKEML is the most complete, but is
            slow for IO).

        .. rubric:: Example

        >>> tribe = Tribe(templates=[Template(name='c', st=read())])
        >>> tribe.write('test_tribe')
        Tribe of 1 templates
        >>> tribe.write(
        ...    "this_wont_work.bob",
        ...    catalog_format="BOB") # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        TypeError: BOB is not supported
        """
        from eqcorrscan.core.match_filter import CAT_EXT_MAP

        if catalog_format not in CAT_EXT_MAP.keys():
            raise TypeError("{0} is not supported".format(catalog_format))
        dirname, ext = os.path.splitext(filename)
        if not os.path.isdir(dirname):
            os.makedirs(dirname)
        self._par_write(dirname)
        tribe_cat = Catalog()
        for t in self.templates:
            if t.event is not None:
                # Check that the name in the comment matches the template name
                for comment in t.event.comments:
                    if not comment.text:
                        comment.text = "eqcorrscan_template_{0}".format(t.name)
                    elif comment.text.startswith("eqcorrscan_template_"):
                        comment.text = "eqcorrscan_template_{0}".format(t.name)
                tribe_cat.append(t.event)
        if len(tribe_cat) > 0:
            tribe_cat.write(
                os.path.join(dirname, 'tribe_cat.{0}'.format(
                    CAT_EXT_MAP[catalog_format])), format=catalog_format)
        for template in self.templates:
            template.st.write(
                os.path.join(dirname, '{0}.ms'.format(template.name)),
                format='MSEED')
        if compress:
            if not filename.endswith(".tgz"):
                Logger.info("Appending '.tgz' to filename.")
                filename += ".tgz"
            with tarfile.open(filename, "w:gz") as tar:
                tar.add(dirname, arcname=os.path.basename(dirname))
            shutil.rmtree(dirname)
        return self

    def _par_write(self, dirname):
        """
        Internal write function to write a formatted parameter file.

        :type dirname: str
        :param dirname: Directory to write the parameter file to.
        """
        filename = dirname + '/' + 'template_parameters.csv'
        with open(filename, 'w') as parfile:
            for template in self.templates:
                for key in template.__dict__.keys():
                    if key not in ['st', 'event']:
                        parfile.write(key + ': ' +
                                      str(template.__dict__[key]) + ', ')
                parfile.write('\n')
        return self

    def _temporary_template_db(self, template_dir: str = None) -> dict:
        """
        Write a temporary template database of pickled templates to disk.

        :param template_dir:
            Directory to write to - if None will make a temporary directory.
        """
        # We use template names for filenames - check that these are unique
        self.__unique_ids()
        # Make sure that the template directory exists, or make a tempdir
        if template_dir:
            if not os.path.isdir(template_dir):
                os.makedirs(template_dir)
        else:
            template_dir = tempfile.mkdtemp()
        template_files = dict()
        for template in self.templates:
            t_file = os.path.join(template_dir, f"{template.name}.pkl")
            with open(t_file, "wb") as f:
                pickle.dump(template, f)
            template_files.update({template.name: t_file})
        return template_files

    def read(self, filename):
        """
        Read a tribe of templates from a tar formatted file.

        :type filename: str
        :param filename: File to read templates from.

        .. rubric:: Example

        >>> tribe = Tribe(templates=[Template(name='c', st=read())])
        >>> tribe.write('test_tribe')
        Tribe of 1 templates
        >>> tribe_back = Tribe().read('test_tribe.tgz')
        >>> tribe_back == tribe
        True
        >>> # This can also read pickled templates
        >>> import pickle
        >>> with open("test_tribe.pkl", "wb") as f:
        ...    pickle.dump(tribe, f)
        >>> tribe_back = Tribe().read("test_tribe.pkl")
        >>> tribe_back == tribe
        True
        """
        if filename.endswith(".pkl"):
            with open(filename, "rb") as f:
                self.__iadd__(pickle.load(f))
            return self
        with tarfile.open(filename, "r:*") as arc:
            temp_dir = tempfile.mkdtemp()
            arc.extractall(path=temp_dir, members=_safemembers(arc))
            tribe_dir = glob.glob(temp_dir + os.sep + '*')[0]
            self._read_from_folder(dirname=tribe_dir)
        shutil.rmtree(temp_dir)
        # Assign unique ids
        self.__unique_ids()
        return self

    def _read_from_folder(self, dirname):
        """
        Internal folder reader.

        :type dirname: str
        :param dirname: Folder to read from.
        """
        templates = _par_read(dirname=dirname, compressed=False)
        t_files = glob.glob(dirname + os.sep + '*.ms')
        tribe_cat_file = glob.glob(os.path.join(dirname, "tribe_cat.*"))
        if len(tribe_cat_file) != 0:
            tribe_cat = read_events(tribe_cat_file[0])
        else:
            tribe_cat = Catalog()
        previous_template_names = [t.name for t in self.templates]
        for template in templates:
            if template.name in previous_template_names:
                # Don't read in for templates that we already have.
                continue
            for event in tribe_cat:
                for comment in event.comments:
                    if comment.text == 'eqcorrscan_template_' + template.name:
                        template.event = event
            t_file = [t for t in t_files
                      if t.split(os.sep)[-1] == template.name + '.ms']
            if len(t_file) == 0:
                Logger.error('No waveform for template: ' + template.name)
                continue
            elif len(t_file) > 1:
                Logger.warning('Multiple waveforms found, using: ' + t_file[0])
            template.st = read(t_file[0])
        # Remove templates that do not have streams
        templates = [t for t in templates if t.st is not None]
        self.templates.extend(templates)
        return

    def construct(self, method, lowcut, highcut, samp_rate, filt_order,
                  length, prepick, swin="all", process_len=86400,
                  all_horiz=False, delayed=True, plot=False, plotdir=None,
                  min_snr=None, parallel=False, num_cores=False,
                  skip_short_chans=False, save_progress=False, **kwargs):
        """
        Generate a Tribe of Templates.

        :type method: str
        :param method:
            Method of Tribe generation. Possible options are: `from_client`,
            `from_meta_file`.  See below on the additional required arguments
            for each method.
        :type lowcut: float
        :param lowcut:
            Low cut (Hz), if set to None will not apply a lowcut
        :type highcut: float
        :param highcut:
            High cut (Hz), if set to None will not apply a highcut.
        :type samp_rate: float
        :param samp_rate:
            New sampling rate in Hz.
        :type filt_order: int
        :param filt_order:
            Filter level (number of corners).
        :type length: float
        :param length: Length of template waveform in seconds.
        :type prepick: float
        :param prepick: Pre-pick time in seconds
        :type swin: str
        :param swin:
            P, S, P_all, S_all or all, defaults to all: see note in
            :func:`eqcorrscan.core.template_gen.template_gen`
        :type process_len: int
        :param process_len: Length of data in seconds to download and process.
        :type all_horiz: bool
        :param all_horiz:
            To use both horizontal channels even if there is only a pick on
            one of them.  Defaults to False.
        :type delayed: bool
        :param delayed: If True, each channel will begin relative to it's own
            pick-time, if set to False, each channel will begin at the same
            time.
        :type plot: bool
        :param plot: Plot templates or not.
        :type plotdir: str
        :param plotdir:
            The path to save plots to. If `plotdir=None` (default) then the
            figure will be shown on screen.
        :type min_snr: float
        :param min_snr:
            Minimum signal-to-noise ratio for a channel to be included in the
            template, where signal-to-noise ratio is calculated as the ratio
            of the maximum amplitude in the template window to the rms
            amplitude in the whole window given.
        :type parallel: bool
        :param parallel: Whether to process data in parallel or not.
        :type num_cores: int
        :param num_cores:
            Number of cores to try and use, if False and parallel=True,
            will use either all your cores, or as many traces as in the data
            (whichever is smaller).
        :type save_progress: bool
        :param save_progress:
            Whether to save the resulting template set at every data step or
            not. Useful for long-running processes.
        :type skip_short_chans: bool
        :param skip_short_chans:
            Whether to ignore channels that have insufficient length data or
            not. Useful when the quality of data is not known, e.g. when
            downloading old, possibly triggered data from a datacentre

        .. note::
            *Method specific arguments:*

            - `from_client` requires:
                :param str client_id:
                    string passable by obspy to generate Client, or any object
                    with a `get_waveforms` method, including a Client instance.
                :param `obspy.core.event.Catalog` catalog:
                    Catalog of events to generate template for
                :param float data_pad: Pad length for data-downloads in seconds
            - `from_meta_file` requires:
                :param str meta_file:
                    Path to obspy-readable event file, or an obspy Catalog
                :param `obspy.core.stream.Stream` st:
                    Stream containing waveform data for template. Note that
                    this should be the same length of stream as you will use
                    for the continuous detection, e.g. if you detect in
                    day-long files, give this a day-long file!
                :param bool process:
                    Whether to process the data or not, defaults to True.

        .. Note::
            Method: `from_sac` is not supported by Tribe.construct and must
            use Template.construct.

        .. Note:: Templates will be named according to their start-time.
        """
        templates, catalog, process_lengths = template_gen.template_gen(
            method=method, lowcut=lowcut, highcut=highcut, length=length,
            filt_order=filt_order, samp_rate=samp_rate, prepick=prepick,
            return_event=True, save_progress=save_progress, swin=swin,
            process_len=process_len, all_horiz=all_horiz, plotdir=plotdir,
            delayed=delayed, plot=plot, min_snr=min_snr, parallel=parallel,
            num_cores=num_cores, skip_short_chans=skip_short_chans,
            **kwargs)
        for template, event, process_len in zip(templates, catalog,
                                                process_lengths):
            t = Template()
            if len(template) == 0:
                Logger.error('Empty Template')
                continue
            t.st = template
            t.name = template.sort(['starttime'])[0]. \
                stats.starttime.strftime('%Y_%m_%dt%H_%M_%S')
            t.lowcut = lowcut
            t.highcut = highcut
            t.filt_order = filt_order
            t.samp_rate = samp_rate
            t.process_length = process_len
            t.prepick = prepick
            event.comments.append(Comment(
                text="eqcorrscan_template_" + t.name,
                creation_info=CreationInfo(agency='eqcorrscan',
                                           author=getpass.getuser())))
            t.event = event
            self.templates.append(t)
        return self

    def _template_channel_ids(self, wildcards: bool = False):
        template_channel_ids = set()
        for template in self.templates:
            for tr in template.st:
                # Cope with missing info and convert to wildcards
                net, sta, loc, chan = tr.id.split('.')
                if wildcards:
                    net, sta, loc, chan = _wildcard_fill(
                        net, sta, loc, chan)
                template_channel_ids.add((net, sta, loc, chan))
        return template_channel_ids

    def cluster(self, method, **kwargs):
        """
        Cluster the tribe.

        Cluster templates within a tribe: returns multiple tribes each of
        which could be stacked.

        :type method: str
        :param method:
            Method of stacking, see :mod:`eqcorrscan.utils.clustering`

        :return: List of tribes.
        """
        from eqcorrscan.utils import clustering
        tribes = []
        func = getattr(clustering, method)
        if method in ['space_cluster', 'space_time_cluster']:
            cat = Catalog([t.event for t in self.templates])
            groups = func(cat, **kwargs)
            for group in groups:
                new_tribe = Tribe()
                for event in group:
                    new_tribe.templates.extend([t for t in self.templates
                                                if t.event == event])
                tribes.append(new_tribe)
        return tribes

    def detect(self, stream, threshold, threshold_type, trig_int, plot=False,
               plotdir=None, daylong=False, parallel_process=True,
               xcorr_func=None, concurrency=None, cores=None,
               concurrent_processing=False, ignore_length=False,
               ignore_bad_data=False, group_size=None, overlap="calculate",
               full_peaks=False, save_progress=False, process_cores=None,
               pre_processed=False, check_processing=True,
               **kwargs):
        """
        Detect using a Tribe of templates within a continuous stream.

        :type stream: `Queue` or `obspy.core.stream.Stream`
        :param stream:
            Queue of streams of continuous data to detect within using the
            Templates, or just the continuous data itself.
        :type threshold: float
        :param threshold:
            Threshold level, if using `threshold_type='MAD'` then this will be
            the multiple of the median absolute deviation.
        :type threshold_type: str
        :param threshold_type:
            The type of threshold to be used, can be MAD, absolute or
            av_chan_corr.  See Note on thresholding below.
        :type trig_int: float
        :param trig_int:
            Minimum gap between detections from one template in seconds.
            If multiple detections occur within trig_int of one-another, the
            one with the highest cross-correlation sum will be selected.
        :type plot: bool
        :param plot: Turn plotting on or off.
        :type plotdir: str
        :param plotdir:
            The path to save plots to. If `plotdir=None` (default) then the
            figure will be shown on screen.
        :type daylong: bool
        :param daylong:
            Set to True to use the
            :func:`eqcorrscan.utils.pre_processing.dayproc` routine, which
            preforms additional checks and is more efficient for day-long data
            over other methods.
        :type parallel_process: bool
        :param parallel_process:
        :type xcorr_func: str or callable
        :param xcorr_func:
            A str of a registered xcorr function or a callable for implementing
            a custom xcorr function. For more information see:
            :func:`eqcorrscan.utils.correlate.register_array_xcorr`
        :type concurrency: str
        :param concurrency:
            The type of concurrency to apply to the xcorr function. Options are
            'multithread', 'multiprocess', 'concurrent'. For more details see
            :func:`eqcorrscan.utils.correlate.get_stream_xcorr`
        :type cores: int
        :param cores: Number of workers for processing and detection.
        :type concurrent_processing: bool
        :param concurrent_processing:
            Whether to process steps in detection workflow concurrently or not.
            See https://github.com/eqcorrscan/EQcorrscan/pull/544 for
            benchmarking.
        :type ignore_length: bool
        :param ignore_length:
            If using daylong=True, then dayproc will try check that the data
            are there for at least 80% of the day, if you don't want this check
            (which will raise an error if too much data are missing) then set
            ignore_length=True.  This is not recommended!
        :type ignore_bad_data: bool
        :param ignore_bad_data:
            If False (default), errors will be raised if data are excessively
            gappy or are mostly zeros. If True then no error will be raised,
            but an empty trace will be returned (and not used in detection).
        :type group_size: int
        :param group_size:
            Maximum number of templates to run at once, use to reduce memory
            consumption, if unset will use all templates.
        :type overlap: float
        :param overlap:
            Either None, "calculate" or a float of number of seconds to
            overlap detection streams by.  This is to counter the effects of
            the delay-and-stack in calculating cross-correlation sums. Setting
            overlap = "calculate" will work out the appropriate overlap based
            on the maximum lags within templates.
        :type full_peaks: bool
        :param full_peaks: See `eqcorrscan.utils.findpeaks.find_peaks2_short`
        :type save_progress: bool
        :param save_progress:
            Whether to save the resulting party at every data step or not.
            Useful for long-running processes.
        :type process_cores: int
        :param process_cores:
            Number of processes to use for pre-processing (if different to
            `cores`).
        :type pre_processed: bool
        :param pre_processed:
            Whether the stream has been pre-processed or not to match the
            templates.
        :type check_processing: bool
        :param check_processing:
            Whether to check that all templates were processed the same.

        :return:
            :class:`eqcorrscan.core.match_filter.Party` of Families of
            detections.

        .. Note::
            When using the "fftw" correlation backend the length of the fft
            can be set. See :mod:`eqcorrscan.utils.correlate` for more info.

        .. Note::
            `stream` must not be pre-processed. If your data contain gaps
            you should *NOT* fill those gaps before using this method.
            The pre-process functions (called within) will fill the gaps
            internally prior to processing, process the data, then re-fill
            the gaps with zeros to ensure correlations are not incorrectly
            calculated within gaps. If your data have gaps you should pass a
            merged stream without the `fill_value` argument
            (e.g.: `stream = stream.merge()`).

        .. note::
            **Data overlap:**

            Internally this routine shifts and trims the data according to the
            offsets in the template (e.g. if trace 2 starts 2 seconds after
            trace 1 in the template then the continuous data will be shifted
            by 2 seconds to align peak correlations prior to summing).
            Because of this, detections at the start and end of continuous
            data streams **may be missed**.  The maximum time-period that
            might be missing detections is the maximum offset in the template.

            To work around this, if you are conducting matched-filter
            detections through long-duration continuous data, we suggest
            using some overlap (a few seconds, on the order of the maximum
            offset in the templates) in the continuous data.  You will then
            need to post-process the detections (which should be done anyway
            to remove duplicates).  See below note for how `overlap` argument
            affects data internally if `stream` is longer than the processing
            length.

        .. Note::
            If `stream` is longer than processing length, this routine will
            ensure that data overlap between loops, which will lead to no
            missed detections at data start-stop points (see above note).
            This will result in end-time not being strictly
            honoured, so detections may occur after the end-time set.  This is
            because data must be run in the correct process-length.

        .. note::
            **Thresholding:**

            **MAD** threshold is calculated as the:

            .. math::

                threshold {\\times} (median(abs(cccsum)))

            where :math:`cccsum` is the cross-correlation sum for a given
            template.

            **absolute** threshold is a true absolute threshold based on the
            cccsum value.

            **av_chan_corr** is based on the mean values of single-channel
            cross-correlations assuming all data are present as required for
            the template, e.g:

            .. math::

                av_chan_corr_thresh=threshold \\times (cccsum /
                len(template))

            where :math:`template` is a single template from the input and the
            length is the number of channels within this template.
        """
        # Check that template names are unique
        self.__unique_ids()
        # We should not need to copy the stream, it is copied in chunks by
        # _group_process

        # Argument handling
        if overlap is None:
            overlap = 0.0
        elif not isinstance(overlap, float) and str(overlap) == "calculate":
            overlap = max(
                _moveout(template.st) for template in self.templates)
        elif not isinstance(overlap, float):
            raise NotImplementedError(
                "%s is not a recognised overlap type" % str(overlap))
        assert overlap < self.templates[0].process_length, (
            f"Overlap {overlap} must be less than process length "
            f"{self.templates[0].process_length}")

        # Copy because we need to muck around with them.
        inner_kwargs = copy.copy(kwargs)

        plot_format = inner_kwargs.pop("plot_format", "png")
        export_cccsums = inner_kwargs.pop('export_cccsums', False)
        peak_cores = inner_kwargs.pop('peak_cores',
                                      process_cores) or cpu_count()
        groups = inner_kwargs.pop("groups", None)

        if peak_cores == 1:
            parallel = False
        else:
            parallel = True

        if check_processing:
            assert len(quick_group_templates(self.templates)) == 1, (
                "Inconsistent template processing parameters found, this is no"
                " longer supported. \nSplit your tribe using "
                "eqcorrscan.core.match_filter.template.quick_group_templates "
                "and re-run for each grouped tribe")
        sampling_rate = self.templates[0].samp_rate
        # Used for sanity checking seed id overlap
        template_ids = set(
            tr.id for template in self.templates for tr in template.st)

        args = (stream, template_ids, pre_processed, parallel_process,
                process_cores, daylong, ignore_length, overlap,
                ignore_bad_data, group_size, groups, sampling_rate, threshold,
                threshold_type, save_progress, xcorr_func, concurrency, cores,
                export_cccsums, parallel, peak_cores, trig_int, full_peaks,
                plot, plotdir, plot_format,)

        if concurrent_processing:
            party = self._detect_concurrent(*args, **inner_kwargs)
        else:
            party = self._detect_serial(*args, **inner_kwargs)

        Logger.info("Ensuring all templates are in party")
        additional_families = []
        for template in self.templates:
            if template.name in party._template_dict.keys():
                continue
            additional_families.append(
                Family(template=template, detections=[]))
        party.families.extend(additional_families)

        # Post-process
        if len(party) > 0:
            Logger.info("Removing duplicates")
            party = _remove_duplicates(party)
        return party

    def _detect_serial(
        self, stream, template_ids, pre_processed, parallel_process,
        process_cores, daylong, ignore_length, overlap, ignore_bad_data,
        group_size, groups, sampling_rate, threshold, threshold_type,
        save_progress, xcorr_func, concurrency, cores, export_cccsums,
        parallel, peak_cores, trig_int, full_peaks, plot, plotdir, plot_format,
        **kwargs
    ):
        """ Internal serial detect workflow. """
        from eqcorrscan.core.match_filter.helpers.tribe import (
            _pre_process, _group, _corr_and_peaks, _detect, _make_party)
        party = Party()

        assert isinstance(stream, Stream), (
            f"Serial detection requires stream to be a stream, not"
            f" a {type(stream)}")

        # We need to copy data here to keep the users input safe.
        st_chunks = _pre_process(
            st=stream.copy(), template_ids=template_ids,
            pre_processed=pre_processed,
            filt_order=self.templates[0].filt_order,
            highcut=self.templates[0].highcut,
            lowcut=self.templates[0].lowcut,
            samp_rate=self.templates[0].samp_rate,
            process_length=self.templates[0].process_length,
            parallel=parallel_process, cores=process_cores, daylong=daylong,
            ignore_length=ignore_length, ignore_bad_data=ignore_bad_data,
            overlap=overlap, **kwargs)

        chunk_files = []
        for st_chunk in st_chunks:
            starttime = st_chunk[0].stats.starttime
            delta = st_chunk[0].stats.delta
            template_groups = _group(
                sids={tr.id for tr in st_chunk},
                templates=self.templates, group_size=group_size, groups=groups)
            for i, template_group in enumerate(template_groups):
                templates = [_quick_copy_stream(t.st) for t in template_group]
                template_names = [t.name for t in template_group]
                Logger.info(
                    f"Prepping {len(templates)} templates for correlation")

                # We need to copy the stream here.
                _st, templates, template_names = _prep_data_for_correlation(
                    stream=_quick_copy_stream(st_chunk), templates=templates,
                    template_names=template_names)

                all_peaks, thresholds, no_chans, chans = _corr_and_peaks(
                    templates=templates, template_names=template_names,
                    stream=_st, xcorr_func=xcorr_func, concurrency=concurrency,
                    cores=cores, i=i, export_cccsums=export_cccsums,
                    parallel=parallel, peak_cores=peak_cores,
                    threshold=threshold, threshold_type=threshold_type,
                    trig_int=trig_int, sampling_rate=sampling_rate,
                    full_peaks=full_peaks, plot=plot, plotdir=plotdir,
                    plot_format=plot_format, prepped=False, **kwargs)

                detections = _detect(
                    template_names=template_names,
                    all_peaks=all_peaks, starttime=starttime,
                    delta=delta, no_chans=no_chans,
                    chans=chans, thresholds=thresholds)

                chunk_file = _make_party(
                    detections=detections, threshold=threshold,
                    threshold_type=threshold_type,
                    templates=self.templates, chunk_start=starttime,
                    chunk_id=i, save_progress=save_progress)
                chunk_files.append(chunk_file)
        # Rebuild
        for _chunk_file in chunk_files:
            Logger.info(f"Adding party from {_chunk_file} to party")
            with open(_chunk_file, "rb") as _f:
                party += pickle.load(_f)
            if not save_progress:
                try:
                    os.remove(_chunk_file)
                except FileNotFoundError:
                    pass
            Logger.info(f"Added party from {_chunk_file}, party now "
                        f"contains {len(party)} detections")

        if os.path.isdir(self._stream_dir):
            shutil.rmtree(self._stream_dir)
        return party

    def _detect_concurrent(
        self, stream, template_ids, pre_processed, parallel_process,
        process_cores, daylong, ignore_length, overlap, ignore_bad_data,
        group_size, groups, sampling_rate, threshold, threshold_type,
        save_progress, xcorr_func, concurrency, cores, export_cccsums,
        parallel, peak_cores, trig_int, full_peaks, plot, plotdir, plot_format,
        **kwargs
    ):
        """ Internal concurrent detect workflow. """
        from eqcorrscan.core.match_filter.helpers.processes import (
            _pre_processor, _prepper, _make_detections, _check_for_poison,
            _get_and_check, Poison)
        from eqcorrscan.core.match_filter.helpers.tribe import _corr_and_peaks

        if isinstance(stream, Stream):
            Logger.info("Copying stream to keep your original data safe")
            st_queue = Queue(maxsize=2)
            st_queue.put(stream.copy())
            # Close off queue
            st_queue.put(None)
            stream = st_queue
        else:
            # Note that if a queue has been passed we do not try to keep
            # data safe
            Logger.warning("Streams in queue will be edited in-place, you "
                           "should not re-use them")

        # To reduce load copying templates between processes we dump them to
        # disk and pass the dictionary of files
        template_dir = f".template_db_{uuid.uuid4()}"
        template_db = self._temporary_template_db(template_dir)

        # Set up processes and queues
        poison_queue = kwargs.get('poison_queue', Queue())

        if not pre_processed:
            processed_stream_queue = Queue(maxsize=1)
        else:
            processed_stream_queue = stream

        # Prepped queue contains templates and stream (and extras)
        prepped_queue = Queue(maxsize=1)
        # Output queues
        peaks_queue = Queue()
        party_file_queue = Queue()

        # Set up processes
        if not pre_processed:
            pre_processor_process = Process(
                target=_pre_processor,
                kwargs=dict(
                    input_stream_queue=stream,
                    temp_stream_dir=self._stream_dir,
                    template_ids=template_ids,
                    pre_processed=pre_processed,
                    filt_order=self.templates[0].filt_order,
                    highcut=self.templates[0].highcut,
                    lowcut=self.templates[0].lowcut,
                    samp_rate=self.templates[0].samp_rate,
                    process_length=self.templates[0].process_length,
                    parallel=parallel_process,
                    cores=process_cores,
                    daylong=daylong,
                    ignore_length=ignore_length,
                    overlap=overlap,
                    ignore_bad_data=ignore_bad_data,
                    output_filename_queue=processed_stream_queue,
                    poison_queue=poison_queue,
                ),
                name="ProcessProcess"
            )

        prepper_process = Process(
            target=_prepper,
            kwargs=dict(
                input_stream_filename_queue=processed_stream_queue,
                group_size=group_size,
                groups=groups,
                templates=template_db,
                output_queue=prepped_queue,
                poison_queue=poison_queue,
                xcorr_func=xcorr_func,
            ),
            name="PrepProcess"
        )
        detector_process = Process(
            target=_make_detections,
            kwargs=dict(
                input_queue=peaks_queue,
                delta=1 / sampling_rate,
                templates=template_db,
                threshold=threshold,
                threshold_type=threshold_type,
                save_progress=save_progress,
                output_queue=party_file_queue,
                poison_queue=poison_queue,
            ),
            name="DetectProcess"
        )

        # Cope with old tribes
        if not hasattr(self, '_processes'):
            self._processes = dict()
        if not hasattr(self, '_queues'):
            self._queues = dict()

        # Put these processes into the namespace
        self._processes.update({
            "prepper": prepper_process,
            "detector": detector_process,
        })
        self._queues.update({
            "poison": poison_queue,
            "stream": stream,
            "prepped": prepped_queue,
            "peaks": peaks_queue,
            "party_file": party_file_queue,
        })

        if not pre_processed:
            Logger.info("Starting preprocessor")
            self._queues.update({"processed_stream": processed_stream_queue})
            self._processes.update({"pre-processor": pre_processor_process})
            pre_processor_process.start()

        # Start your engines!
        prepper_process.start()
        detector_process.start()

        # Loop over input streams and template groups
        while True:
            killed = _check_for_poison(poison_queue)
            if killed:
                Logger.info("Killed in main loop")
                break
            try:
                to_corr = _get_and_check(prepped_queue, poison_queue)
                if to_corr is None:
                    Logger.info("Ran out of streams, exiting correlation")
                    break
                elif isinstance(to_corr, Poison):
                    Logger.info("Killed in main loop")
                    break
                starttime, i, stream, template_names, templates, \
                    *extras = to_corr
                inner_kwargs = copy.copy(kwargs)  # We will mangle them
                # Correlation specific handling to reduce single-threaded time
                if xcorr_func == "fmf":
                    weights, pads, chans, no_chans = extras
                    inner_kwargs.update({
                        'weights': weights, 'pads': pads, "no_chans": no_chans,
                        "chans": chans, "prepped": True})
                elif xcorr_func in (None, 'fftw'):
                    pads, seed_ids = extras
                    inner_kwargs.update({
                        "pads": pads, "seed_ids": seed_ids, "prepped": True})

                Logger.info(f"Got stream of {len(stream)} channels")
                Logger.info(f"Starting correlation from {starttime}")

                all_peaks, thresholds, no_chans, chans = _corr_and_peaks(
                    templates=templates, template_names=template_names,
                    stream=stream, xcorr_func=xcorr_func,
                    concurrency=concurrency,
                    cores=cores, i=i, export_cccsums=export_cccsums,
                    parallel=parallel, peak_cores=peak_cores,
                    threshold=threshold, threshold_type=threshold_type,
                    trig_int=trig_int, sampling_rate=sampling_rate,
                    full_peaks=full_peaks, plot=plot, plotdir=plotdir,
                    plot_format=plot_format, **inner_kwargs
                )
                peaks_queue.put(
                    (starttime, all_peaks, thresholds, no_chans, chans,
                     template_names))
            except Exception as e:
                Logger.error(
                    f"Caught exception in correlator:\n {e}")
                traceback.print_tb(e.__traceback__)
                poison_queue.put(e)
                break  # We need to break in Main
            i += 1
        Logger.debug("Putting None into peaks queue.")
        peaks_queue.put(None)

        # Get the party back
        Logger.info("Collecting party")
        party = Party()
        while True:
            killed = _check_for_poison(poison_queue)
            if killed:
                Logger.error("Killed")
                break
            pf = _get_and_check(party_file_queue, poison_queue)
            if pf is None:
                # Fin - Queue has been finished with.
                break
            if isinstance(pf, Poison):
                Logger.error("Killed while checking for party")
                killed = True
                break
            with open(pf, "rb") as f:
                party += pickle.load(f)
            if not save_progress:
                try:
                    os.remove(pf)
                except FileNotFoundError:
                    pass

        # Check for exceptions
        if killed:
            internal_error = poison_queue.get()
            if isinstance(internal_error, Poison):
                internal_error = internal_error.value
            Logger.error(f"Raising error {internal_error} in main process")
            # Now we can raise the error
            if internal_error:
                # Clean the template db
                if os.path.isdir(template_dir):
                    shutil.rmtree(template_dir)
                if os.path.isdir(self._stream_dir):
                    shutil.rmtree(self._stream_dir)
                self._on_error(internal_error)

        # Shut down the processes and close the queues
        shutdown = kwargs.get("shutdown", True)
        # Allow client_detect to take control
        if shutdown:
            Logger.info("Shutting down")
            self._close_queues()
            self._close_processes()
        if os.path.isdir(template_dir):
            shutil.rmtree(template_dir)
        if os.path.isdir(self._stream_dir):
            shutil.rmtree(self._stream_dir)
        return party

    def client_detect(self, client, starttime, endtime, threshold,
                      threshold_type, trig_int, plot=False, plotdir=None,
                      min_gap=None, daylong=False, parallel_process=True,
                      xcorr_func=None, concurrency=None, cores=None,
                      concurrent_processing=False, ignore_length=False,
                      ignore_bad_data=False, group_size=None,
                      return_stream=False, full_peaks=False,
                      save_progress=False, process_cores=None, retries=3,
                      check_processing=True, **kwargs):
        """
        Detect using a Tribe of templates within a continuous stream.

        :type client: `obspy.clients.*.Client`
        :param client:
            Any obspy client (or client-like object) with a dataselect service.
        :type starttime: :class:`obspy.core.UTCDateTime`
        :param starttime: Start-time for detections.
        :type endtime: :class:`obspy.core.UTCDateTime`
        :param endtime: End-time for detections
        :type threshold: float
        :param threshold:
            Threshold level, if using `threshold_type='MAD'` then this will be
            the multiple of the median absolute deviation.
        :type threshold_type: str
        :param threshold_type:
            The type of threshold to be used, can be MAD, absolute or
            av_chan_corr.  See Note on thresholding below.
        :type trig_int: float
        :param trig_int:
            Minimum gap between detections from one template in seconds.
            If multiple detections occur within trig_int of one-another, the
            one with the highest cross-correlation sum will be selected.
        :type plot: bool
        :param plot: Turn plotting on or off.
        :type plotdir: str
        :param plotdir:
            The path to save plots to. If `plotdir=None` (default) then the
            figure will be shown on screen.
        :type min_gap: float
        :param min_gap:
            Minimum gap allowed in data - use to remove traces with known
            issues
        :type daylong: bool
        :param daylong:
            Set to True to use the
            :func:`eqcorrscan.utils.pre_processing.dayproc` routine, which
            preforms additional checks and is more efficient for day-long data
            over other methods.
        :type parallel_process: bool
        :param parallel_process:
        :type xcorr_func: str or callable
        :param xcorr_func:
            A str of a registered xcorr function or a callable for implementing
            a custom xcorr function. For more information see:
            :func:`eqcorrscan.utils.correlate.register_array_xcorr`
        :type concurrency: str
        :param concurrency:
            The type of concurrency to apply to the xcorr function. Options are
            'multithread', 'multiprocess', 'concurrent'. For more details see
            :func:`eqcorrscan.utils.correlate.get_stream_xcorr`
        :type cores: int
        :param cores: Number of workers for processing and detection.
        :type concurrent_processing: bool
        :param concurrent_processing:
            Whether to process steps in detection workflow concurrently or not.
            See https://github.com/eqcorrscan/EQcorrscan/pull/544 for
            benchmarking.
        :type ignore_length: bool
        :param ignore_length:
            If using daylong=True, then dayproc will try check that the data
            are there for at least 80% of the day, if you don't want this check
            (which will raise an error if too much data are missing) then set
            ignore_length=True.  This is not recommended!
        :type ignore_bad_data: bool
        :param ignore_bad_data:
            If False (default), errors will be raised if data are excessively
            gappy or are mostly zeros. If True then no error will be raised,
            but an empty trace will be returned (and not used in detection).
        :type group_size: int
        :param group_size:
            Maximum number of templates to run at once, use to reduce memory
            consumption, if unset will use all templates.
        :type full_peaks: bool
        :param full_peaks: See `eqcorrscan.utils.findpeaks.find_peaks2_short`
        :type save_progress: bool
        :param save_progress:
            Whether to save the resulting party at every data step or not.
            Useful for long-running processes.
        :type process_cores: int
        :param process_cores:
            Number of processes to use for pre-processing (if different to
            `cores`).
        :type return_stream: bool
        :param return_stream:
            Whether to also output the stream downloaded, useful if you plan
            to use the stream for something else, e.g. lag_calc.
        :type retries: int
        :param retries:
            Number of attempts allowed for downloading - allows for transient
            server issues.

        :return:
            :class:`eqcorrscan.core.match_filter.Party` of Families of
            detections.


        .. Note::
            When using the "fftw" correlation backend the length of the fft
            can be set. See :mod:`eqcorrscan.utils.correlate` for more info.

        .. Note::
            Ensures that data overlap between loops, which will lead to no
            missed detections at data start-stop points (see note for
            :meth:`eqcorrscan.core.match_filter.Tribe.detect` method).
            This will result in end-time not being strictly
            honoured, so detections may occur after the end-time set.  This is
            because data must be run in the correct process-length.

        .. warning::
            Plotting within the match-filter routine uses the Agg backend
            with interactive plotting turned off.  This is because the function
            is designed to work in bulk.  If you wish to turn interactive
            plotting on you must import matplotlib in your script first,
            when you then import match_filter you will get the warning that
            this call to matplotlib has no effect, which will mean that
            match_filter has not changed the plotting behaviour.

        .. note::
            **Thresholding:**

            **MAD** threshold is calculated as the:

            .. math::

                threshold {\\times} (median(abs(cccsum)))

            where :math:`cccsum` is the cross-correlation sum for a given
            template.

            **absolute** threshold is a true absolute threshold based on the
            cccsum value.

            **av_chan_corr** is based on the mean values of single-channel
            cross-correlations assuming all data are present as required for
            the template, e.g:

            .. math::

                av_chan_corr_thresh=threshold \\times (cccsum /
                len(template))

            where :math:`template` is a single template from the input and the
            length is the number of channels within this template.
        """
        from eqcorrscan.core.match_filter.helpers.tribe import _download_st
        from eqcorrscan.core.match_filter.helpers.processes import (
            _get_detection_stream, _FDSN_parseable_Client)

        # This uses get_waveforms_bulk to get data - not all client types have
        # this, so we check and monkey patch here.
        if not hasattr(client, "get_waveforms_bulk"):
            assert hasattr(client, "get_waveforms"), (
                f"client {client} must have at least a get_waveforms method")
            Logger.info(f"Client {client} does not have a get_waveforms_bulk "
                        "method, monkey-patching this")
            client = get_waveform_client(client)

        if check_processing:
            assert len(quick_group_templates(self.templates)) == 1, (
                "Inconsistent template processing parameters found, this is no"
                " longer supported. Split your tribe using "
                "eqcorrscan.core.match_filter.template.quick_group_templates "
                "and re-run for each group")

        groups = kwargs.get("groups", None)

        # Hard-coded buffer for downloading data, often data downloaded is
        # not the correct length
        buff = 300
        data_length = max([t.process_length for t in self.templates])

        # Calculate overlap
        overlap = max(_moveout(template.st) for template in self.templates)
        assert overlap < data_length, (
            f"Overlap of {overlap} s is larger than the length of data to "
            f"be downloaded: {data_length} s - this won't work.")

        # Work out start and end times of chunks to download
        chunk_start, time_chunks = starttime, []
        while chunk_start < endtime:
            time_chunks.append((chunk_start, chunk_start + data_length + 20))
            chunk_start += data_length - overlap

        full_stream_dir = None
        if return_stream:
            full_stream_dir = tempfile.mkdtemp()

        poison_queue = Queue()

        detector_kwargs = dict(
            threshold=threshold, threshold_type=threshold_type,
            trig_int=trig_int, plot=plot, plotdir=plotdir,
            daylong=daylong, parallel_process=parallel_process,
            xcorr_func=xcorr_func, concurrency=concurrency, cores=cores,
            ignore_length=ignore_length, ignore_bad_data=ignore_bad_data,
            group_size=group_size, overlap=None, full_peaks=full_peaks,
            process_cores=process_cores, save_progress=save_progress,
            return_stream=return_stream, check_processing=False,
            poison_queue=poison_queue, shutdown=False,
            concurrent_processing=concurrent_processing, groups=groups)

        if not concurrent_processing:
            Logger.warning("Using concurrent_processing=True can be faster if"
                           "downloading your data takes a long time. See "
                           "https://github.com/eqcorrscan/EQcorrscan/pull/544"
                           "for benchmarks.")
            party = Party()
            if return_stream:
                full_st = Stream()
            for _starttime, _endtime in time_chunks:
                st = _download_st(
                    starttime=_starttime, endtime=_endtime,
                    buff=buff, min_gap=min_gap,
                    template_channel_ids=self._template_channel_ids(),
                    client=client, retries=retries)
                if len(st) == 0:
                    Logger.warning(f"No suitable data between {_starttime} "
                                   f"and {_endtime}, skipping")
                    continue
                party += self.detect(stream=st, pre_processed=False,
                                     **detector_kwargs)
                if return_stream:
                    full_st += st
            if return_stream:
                return party, full_st
            return party

        # Get data in advance
        time_queue = Queue()
        stream_queue = Queue(maxsize=1)

        # Convert client to pickleable client if needed - issue with
        # Python 3.12 Windows and MacOS
        if isinstance(client, FDSNClient):
            client = _FDSN_parseable_Client.from_client(client)

        downloader = Process(
            target=_get_detection_stream,
            kwargs=dict(
                input_time_queue=time_queue,
                client=client,
                retries=retries,
                min_gap=min_gap,
                buff=buff,
                output_filename_queue=stream_queue,
                poison_queue=poison_queue,
                temp_stream_dir=self._stream_dir,
                full_stream_dir=full_stream_dir,
                pre_process=True, parallel_process=parallel_process,
                process_cores=process_cores, daylong=daylong,
                overlap=0.0, ignore_length=ignore_length,
                ignore_bad_data=ignore_bad_data,
                filt_order=self.templates[0].filt_order,
                highcut=self.templates[0].highcut,
                lowcut=self.templates[0].lowcut,
                samp_rate=self.templates[0].samp_rate,
                process_length=self.templates[0].process_length,
                template_channel_ids=self._template_channel_ids(),
            ),
            name="DownloadProcess"
        )

        # Cope with old tribes
        if not hasattr(self, '_processes'):
            self._processes = dict()
        if not hasattr(self, '_queues'):
            self._queues = dict()
        # Put processes and queues into shared state
        self._processes.update({
            "downloader": downloader,
        })
        self._queues.update({
            "times": time_queue,
            "poison": poison_queue,
            "stream": stream_queue,
        })

        # Fill time queue
        for time_chunk in time_chunks:
            time_queue.put(time_chunk)
        # Close off queue
        time_queue.put(None)

        # Start up processes
        downloader.start()

        party = self.detect(
            stream=stream_queue, pre_processed=True, **detector_kwargs)

        # Close and join processes
        self._close_processes()
        self._close_queues()

        if return_stream:
            full_st = read(os.path.join(full_stream_dir, "*"))
            shutil.rmtree(full_stream_dir)
            return party, full_st
        return party

    def _close_processes(
        self,
        terminate: bool = False,
        processes: dict = None
    ):
        processes = processes or self._processes
        for p_name, p in processes.items():
            if terminate:
                Logger.warning(f"Terminating {p_name}")
                p.terminate()
                continue
            try:
                Logger.info(f"Joining {p_name}")
                p.join(timeout=self._timeout)
            except Exception as e:
                Logger.error(f"Failed to join due to {e}: terminating")
                p.terminate()
            Logger.info(f"Closing {p_name}")
            try:
                p.close()
            except Exception as e:
                Logger.error(
                    f"Failed to close {p_name} due to {e}, terminating")
                p.terminate()
        Logger.info("Finished closing processes")
        return

    def _close_queues(self, queues: dict = None):
        queues = queues or self._queues
        for q_name, q in queues.items():
            if q._closed:
                continue
            Logger.info(f"Emptying {q_name}")
            while True:
                try:
                    q.get_nowait()
                except Empty:
                    break
            Logger.info(f"Closing {q_name}")
            q.close()
        Logger.info("Finished closing queues")
        return

    def _on_error(self, error):
        """ Gracefully close all child processes and queues and raise error """
        self._close_processes(terminate=True)
        self._close_queues()
        Logger.info("Admin complete, raising error")
        raise error


def read_tribe(fname):
    """
    Read a Tribe of templates from a tar archive.

    :param fname: Filename to read from
    :return: :class:`eqcorrscan.core.match_filter.Tribe`
    """
    tribe = Tribe()
    tribe.read(filename=fname)
    return tribe


if __name__ == "__main__":
    import doctest

    doctest.testmod()
    # List files to be removed after doctest
    cleanup = ['test_tribe.tgz', "test_tribe.pkl"]
    for f in cleanup:
        if os.path.isfile(f):
            os.remove(f)
        elif os.path.isdir(f):
            shutil.rmtree(f)



"""
Helper functions for common handling tasks for catalog objects.

.. note:: These functions are tools to aid simplification of general scripts,
    they do not cover all use cases, however if you have a use case you want
    to see here, then let the authors know, or implement it yourself and
    contribute it back to the project, or, if its really good, give it to the
    obspy guys!

:copyright:
    EQcorrscan developers.

:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import logging

from collections import Counter
from obspy.core.event import Catalog


Logger = logging.getLogger(__name__)


def filter_picks(catalog, stations=None, channels=None, networks=None,
                 locations=None, top_n_picks=None, evaluation_mode='all',
                 phase_hints=None, enforce_single_pick=False):
    """
    Filter events in the catalog based on a number of parameters.

    :param catalog: Catalog to filter.
    :type catalog: obspy.core.event.Catalog
    :param stations: List for stations to keep picks from.
    :type stations: list
    :param channels: List of channels to keep picks from.
    :type channels: list
    :param networks: List of networks to keep picks from.
    :type networks: list
    :param locations: List of location codes to use
    :type locations: list
    :param top_n_picks: Filter only the top N most used station-channel pairs.
    :type top_n_picks: int
    :param evaluation_mode:
        To select only manual or automatic picks, or use all (default).
    :type evaluation_mode: str
    :param phase_hints: List of retained phase hints, or None to use all
    :type phase_hints: list
    :param enforce_single_pick:
        Method to enforce using only one pick of each phase-hint per
        station or False to leave all. Can be {False, "earliest"}
    :type enforce_single_pick: str


    :return:
        Filtered Catalog - if events are left with no picks, they are removed
        from the catalog.
    :rtype: obspy.core.event.Catalog

    .. note::
        Will filter first by station, then by channel, then by network, if
        using top_n_picks, this will be done last, after the other filters
        have been applied.

    .. note::
        Doesn't work in place on the catalog, your input catalog will be safe
        unless you overwrite it.

    .. note:: Doesn't expand wildcard characters.

    .. rubric:: Example

    >>> from obspy.clients.fdsn import Client
    >>> from eqcorrscan.utils.catalog_utils import filter_picks
    >>> from obspy import UTCDateTime
    >>> client = Client('NCEDC')
    >>> t1 = UTCDateTime(2004, 9, 28)
    >>> t2 = t1 + 86400
    >>> catalog = client.get_events(starttime=t1, endtime=t2, minmagnitude=3,
    ...                             minlatitude=35.7, maxlatitude=36.1,
    ...                             minlongitude=-120.6, maxlongitude=-120.2,
    ...                             includearrivals=True)
    >>> print(len(catalog))
    12
    >>> filtered_catalog = filter_picks(catalog, stations=['BMS', 'BAP',
    ...                                                    'PAG', 'PAN',
    ...                                                    'PBI', 'PKY',
    ...                                                    'YEG', 'WOF'])
    >>> print(len(filtered_catalog))
    12
    >>> stations = []
    >>> for event in filtered_catalog:
    ...     for pick in event.picks:
    ...         stations.append(pick.waveform_id.station_code)
    >>> print(sorted(list(set(stations))))
    ['BAP', 'BMS', 'PAG', 'PAN', 'PBI', 'PKY', 'WOF', 'YEG']
    """
    assert enforce_single_pick in {False, "earliest"}, \
        f"enforce_single_pick={enforce_single_pick} unknown"
    # Don't work in place on the catalog
    filtered_catalog = catalog.copy()

    if stations:
        for event in filtered_catalog:
            if len(event.picks) == 0:
                continue
            event.picks = [pick for pick in event.picks
                           if pick.waveform_id.station_code in stations]
    if channels:
        for event in filtered_catalog:
            if len(event.picks) == 0:
                continue
            event.picks = [pick for pick in event.picks
                           if pick.waveform_id.channel_code in channels]
    if networks:
        for event in filtered_catalog:
            if len(event.picks) == 0:
                continue
            event.picks = [pick for pick in event.picks
                           if pick.waveform_id.network_code in networks]
    if locations:
        for event in filtered_catalog:
            if len(event.picks) == 0:
                continue
            event.picks = [pick for pick in event.picks
                           if pick.waveform_id.location_code in locations]
    if phase_hints:
        for event in filtered_catalog:
            if len(event.picks) == 0:
                continue
            event.picks = [pick for pick in event.picks
                           if pick.phase_hint in phase_hints]

    if evaluation_mode == 'manual':
        for event in filtered_catalog:
            event.picks = [pick for pick in event.picks
                           if pick.evaluation_mode == 'manual']
    elif evaluation_mode == 'automatic':
        for event in filtered_catalog:
            event.picks = [pick for pick in event.picks
                           if pick.evaluation_mode == 'automatic']
    elif evaluation_mode != 'all':
        Logger.warning(
            'Unrecognised evaluation_mode: {0}, using all picks'.format(
                evaluation_mode))
    if top_n_picks:
        all_picks = []
        for event in filtered_catalog:
            all_picks += [(pick.waveform_id.station_code,
                           pick.waveform_id.channel_code)
                          for pick in event.picks]
        counted = Counter(all_picks).most_common()
        all_picks = []
        # Hack around sorting the counter object: Py 2 does it differently to 3
        for i in range(counted[0][1]):
            highest = [item[0] for item in counted
                       if item[1] >= counted[0][1] - i]
            # Sort them by alphabetical order in station
            highest = sorted(highest, key=lambda tup: tup[0])
            for stachan in highest:
                if stachan not in all_picks:
                    all_picks.append(stachan)
            if len(all_picks) > top_n_picks:
                all_picks = all_picks[0:top_n_picks]
                break
        for event in filtered_catalog:
            if len(event.picks) == 0:
                continue
            event.picks = [pick for pick in event.picks
                           if (pick.waveform_id.station_code,
                               pick.waveform_id.channel_code) in all_picks]
    # Remove events without picks
    tmp_catalog = Catalog()
    for event in filtered_catalog:
        if len(event.picks) > 0:
            tmp_catalog.append(event)

    # Finally remove extra picks
    if enforce_single_pick:
        reverse = False
        # TODO: Allow other options
        for ev in tmp_catalog:
            retained_picks = []
            stations = {p.waveform_id.station_code for p in ev.picks}
            for station in stations:
                phase_hints = {p.phase_hint for p in ev.picks
                               if p.waveform_id.station_code == station}
                for phase_hint in phase_hints:
                    picks = [p for p in ev.picks
                             if p.waveform_id.station_code == station
                             and p.phase_hint == phase_hint]
                    picks.sort(key=lambda p: p.time, reverse=reverse)
                    retained_picks.append(picks[0])
            ev.picks = retained_picks

    return tmp_catalog


def spatial_clip(catalog, corners, mindepth=None, maxdepth=None):
    """
    Clip the catalog to a spatial box, can be irregular.

    Can only be irregular in 2D, depth must be between bounds.

    :type catalog: :class:`obspy.core.catalog.Catalog`
    :param catalog: Catalog to clip.
    :type corners: :class:`matplotlib.path.Path`
    :param corners: Corners to clip the catalog to
    :type mindepth: float
    :param mindepth: Minimum depth for earthquakes in km.
    :type maxdepth: float
    :param maxdepth: Maximum depth for earthquakes in km.

    .. Note::
        Corners is expected to be a :class:`matplotlib.path.Path` in the form
        of tuples of (lat, lon) in decimal degrees.
    """
    cat_out = catalog.copy()
    if mindepth is not None:
        for event in cat_out:
            try:
                origin = _get_origin(event)
            except IOError:
                continue
            if origin.depth < mindepth * 1000:
                cat_out.events.remove(event)
    if maxdepth is not None:
        for event in cat_out:
            try:
                origin = _get_origin(event)
            except IOError:
                continue
            if origin.depth > maxdepth * 1000:
                cat_out.events.remove(event)
    for event in cat_out:
        try:
            origin = _get_origin(event)
        except IOError:
            continue
        if not corners.contains_point((origin.latitude, origin.longitude)):
            cat_out.events.remove(event)
    return cat_out


def _get_origin(event):
    """
    Get the origin of an event.

    :type event: :class:`obspy.core.event.Event`
    :param event: Event to get the origin of.
    :return: :class:`obspy.core.event.Origin`
    """
    if event.preferred_origin() is not None:
        origin = event.preferred_origin()
    elif len(event.origins) > 0:
        origin = event.origins[0]
    else:
        raise IndexError('No origin set, cannot constrain')
    return origin


def get_ordered_trace_indices(stream, event, sort_by="distance"):
    """
    Sort the traces of a template by hypocentral distance.

    :type stream: obspy.core.stream.Stream
    :param stream: Stream to sort
    :type event: obspy.core.event.event.Event
    :param event:
        Event object. Only if the event contains picks, and an origin and asso-
        ciated arrivals (for sort_by="distance"), then it can sort the traces.
    :type sort_by: string
    :param sort_by: "distance" (default) or "pick_time"

    :return:
        List of indices ordered according to the traces in the sorted Stream.
    :rtype: List

    .. note::
        Indices of traces without associated sortable information will appear
        last in returned order of indices. If event doesn't contain enough
        information, None will be returned.
        For sort_by="pick_time", the earliest pick time at the station is rele-
        vant, independant of channel. To sort a template-stream by picktime,
        use stream.sort(['starttime']) instead.
    """
    from operator import itemgetter
    from obspy import UTCDateTime

    if not event:
        Logger.warning('No event information found to sort stream.')
        return None
    if not event.picks:
        Logger.warning('No picks found to sort stream.')
        return None
    origin = _get_origin(event)
    value_list = [None for i in range(len(stream))]
    if sort_by == "distance":
        for j, tr in enumerate(stream):
            # Don't change stats.distance if it is set already
            if hasattr(tr.stats, 'distance'):
                if tr.stats.distance is not None:
                    value_list[j] = tr.stats.distance
                    continue
            # Set default distance to 999 degrees.
            value_list[j] = 999
            # find the arrival that matches the trace
            for arrival in origin.arrivals:
                pick = arrival.pick_id.get_referred_object()
                if tr.stats.station == pick.waveform_id.station_code:
                    if arrival.distance is not None:
                        value_list[j] = arrival.distance
                        break
    elif sort_by == "pick_time":
        for j, tr in enumerate(stream):
            value_list[j] = UTCDateTime(9999, 12, 31, 23, 59, 59)
            for pick in event.picks:
                if tr.stats.station == pick.waveform_id.station_code and\
                    (tr.stats.channel == pick.waveform_id.channel_code or
                     (tr.id[-1] in 'NE12' and pick.waveform_id.id in 'NE12' and
                      tr.id[0:-1] == pick.waveform_id.id[0:-1])):
                    if pick.time < value_list[j]:
                        value_list[j] = pick.time
    else:
        raise NotImplementedError('Sorting by ' + sort_by + 'not implemented.')
    trace_indices_ordered, values_sorted = zip(*sorted(enumerate(value_list),
                                                       key=itemgetter(1)))
    return trace_indices_ordered


if __name__ == "__main__":
    import doctest
    doctest.testmod()



"""
Helper functions for seismic data clients.

:copyright:
    EQcorrscan developers.

:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import logging

from obspy import Stream


Logger = logging.getLogger(__name__)


def get_waveform_client(waveform_client):
    """
    Bind a `get_waveforms_bulk` method to client if it doesn't already have one

    :param waveform_client: Obspy client with a `get_waveforms` method

    :returns: waveform_client with `get_waveforms_bulk`.
    """
    def _get_waveforms_bulk_naive(self, bulk_arg):
        """ naive implementation of get_waveforms_bulk that uses iteration. """
        st = Stream()
        for arg in bulk_arg:
            st += self.get_waveforms(*arg)
        return st

    # add waveform_bulk method dynamically if it doesn't exist already
    if not hasattr(waveform_client, "get_waveforms_bulk"):
        bound_method = _get_waveforms_bulk_naive.__get__(waveform_client)
        setattr(waveform_client, "get_waveforms_bulk", bound_method)

    return waveform_client


if __name__ == "__main__":
    import doctest

    doctest.testmod()



"""
Functions to test the functions within the eqcorrscan.utils.catalog_utils \
submodule.
"""
import unittest
import pytest
from obspy.clients.fdsn import Client
from obspy import UTCDateTime

from eqcorrscan.utils.catalog_utils import filter_picks


@pytest.mark.network
class CatalogUtilsTests(unittest.TestCase):
    @classmethod
    @pytest.mark.flaky(reruns=2)  # Rerun the test in case of network timeout
    def setUpClass(cls):
        client = Client(str("NCEDC"))
        t1 = UTCDateTime(2004, 9, 28)
        t2 = t1 + 86400
        cls.catalog = client.get_events(
            starttime=t1, endtime=t2, minmagnitude=3,
            minlatitude=35.7, maxlatitude=36.1,
            minlongitude=-120.6, maxlongitude=-120.2,
            includearrivals=True)
        # Phase hints are not included in the picks, but are in the arrivals
        for ev in cls.catalog:
            for arr in ev.origins[-1].arrivals:
                pick = arr.pick_id.get_referred_object()
                pick.phase_hint = arr.phase

    def test_filter_picks(self):
        """ Test various methods of filtering picks in a catalog."""

        stations = ['BMS', 'BAP', 'PAG', 'PAN', 'PBI', 'PKY', 'YEG', 'WOF']
        channels = ['SHZ', 'SHN', 'SHE', 'SH1', 'SH2']
        networks = ['NC']
        locations = ['']
        top_n_picks = 5
        filtered_catalog = filter_picks(
            catalog=self.catalog.copy(), stations=stations,
            channels=channels, networks=networks,
            locations=locations, top_n_picks=top_n_picks)
        for event in filtered_catalog:
            for pick in event.picks:
                self.assertTrue(pick.waveform_id.station_code in stations)
                self.assertTrue(pick.waveform_id.channel_code in channels)
                self.assertTrue(pick.waveform_id.network_code in networks)
                self.assertTrue(pick.waveform_id.location_code in locations)
        filtered_catalog = filter_picks(catalog=self.catalog.copy(),
                                        top_n_picks=top_n_picks)
        filtered_stations = []
        for event in filtered_catalog:
            for pick in event.picks:
                filtered_stations.append(pick.waveform_id.station_code)
        self.assertEqual(len(list(set(filtered_stations))), top_n_picks)

    def test_filter_phase_hints(self):
        filtered_catalog = filter_picks(
            self.catalog.copy(), phase_hints=["P"])

        phase_hints = set(p.phase_hint for ev in filtered_catalog
                          for p in ev.picks)
        print(phase_hints)
        self.assertEqual(phase_hints, set("P"))

    def test_filter_single_pick(self):
        filtered_catalog = filter_picks(
            self.catalog.copy(), enforce_single_pick="earliest")

        for ev in filtered_catalog:
            stations = {p.waveform_id.station_code for p in ev.picks}
            for station in stations:
                picks = [p for p in ev.picks
                         if p.waveform_id.station_code == station]
                phase_hints = {p.phase_hint for p in picks}
                for phase_hint in phase_hints:
                    matched_picks = [
                        p for p in picks if p.phase_hint == phase_hint]
                    if len(matched_picks) != 1:
                        print(f"Multiple picks for {station} - {phase_hint}")
                        for pick in matched_picks:
                            print(pick)
                    self.assertEqual(1, len(matched_picks))
        return


if __name__ == '__main__':
    unittest.main()



"""
Functions for testing the tutorials - written as a somewhat monolithic test
because we need certain behaviour.
"""
import unittest
import os
import pytest

from eqcorrscan.tutorials import subspace


class TestTutorialScripts(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.testing_path = os.path.join(
            os.path.abspath(os.path.dirname(__file__)), 'test_data')

    @pytest.mark.superslow
    def test_subspace(self):
        """Test the subspace tutorial."""
        print("Running subspace")
        detections = subspace.run_tutorial(plot=False, cores=1, verbose=True)
        print("Subspace ran")
        if not len(detections) == 8:
            for detection in detections:
                print(detection)
        self.assertEqual(len(detections), 8)


if __name__ == '__main__':
    """
    Run tutorial tests
    """
    unittest.main()



"""
Correlation functions for multi-channel cross-correlation of seismic data.

Various routines used mostly for testing, including links to a compiled
routine using FFTW, a Numpy fft routine which uses bottleneck for normalisation
and a compiled time-domain routine. These have varying levels of efficiency,
both in terms of overall speed, and in memory usage.  The time-domain is the
most memory efficient but slowest routine (although fast for small cases of
less than a few hundred correlations), the Numpy routine is fast, but memory
inefficient due to a need to store large double-precision arrays for
normalisation.  The fftw compiled routine is faster and more memory efficient
than the Numpy routine.

:copyright:
    EQcorrscan developers.

:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import contextlib
import copy
import ctypes
import os
import logging
from multiprocessing import Pool as ProcessPool, cpu_count
from multiprocessing.pool import ThreadPool

import numpy as np
import math

from packaging import version
from timeit import default_timer

from obspy import UTCDateTime

from eqcorrscan.utils.libnames import _load_cdll
from eqcorrscan.utils import FMF_INSTALLED
from eqcorrscan.utils.pre_processing import _stream_quick_select


Logger = logging.getLogger(__name__)

# This is for building docs on readthedocs, which has an old version of
# scipy - without this, this module cannot be imported, which breaks the docs
# Instead we define a dummy function that returns what it is given.
READ_THE_DOCS = os.environ.get('READTHEDOCS', None) == 'True'
if not READ_THE_DOCS:
    from scipy.fftpack.helper import next_fast_len
else:
    def next_fast_len(a):
        return a

XCOR_FUNCS = {}  # cache of functions for doing cross correlations

# methods added to each xcorr func registered
# these implement the stream interface
XCORR_STREAM_METHODS = ('multithread', 'multiprocess', 'concurrent',
                        'stream_xcorr')

# these implement the array interface
XCOR_ARRAY_METHODS = ('array_xcorr')

# Gain shift for low-variance stabilization
MULTIPLIER = 1e8

# Minimum version for compatible correlations from Fast Matched Filter
MIN_FMF_VERSION = version.parse("1.4.0")


class CorrelationError(Exception):
    """ Error handling for correlation functions. """

    def __init__(self, value):
        """ Raise error.

        .. rubric:: Example

        >>> CorrelationError("Something happened")
        Something happened
        """
        self.value = value

    def __repr__(self):
        """ Print error value.

        .. rubric:: Example

        >>> print(CorrelationError("Error").__repr__())
        Error
        """
        return self.value

    def __str__(self):
        """ Print otherwise

        .. rubric:: Example

        >>> print(CorrelationError("Error"))
        Error
        """
        return self.value


# ------------------ Context manager for switching out default
class _Context:
    """ class for permanently or temporarily changing items in a dict """

    def __init__(self, cache, value_to_switch):
        """
        :type cache: dict
        :param cache: A dict to store values in
        :type value_to_switch: str
        :param value_to_switch:
            The key in cache to switch based on different contexts
        """
        self.cache = cache
        self.value_to_switch = value_to_switch
        self.previous_value = None

    def __call__(self, new_value, *args, **kwargs):
        """ # TODO change docs if this ever becomes general use
        Set a new value for the default xcorr function.

        This function can be called directly to permanently change the
        default normxcorr function or it may be used as a context manager
        to only modify it temporarily.

        :param new_value:
        :return:
        """
        self.previous_value = copy.deepcopy(
            self.cache.get(self.value_to_switch))
        self.cache[self.value_to_switch] = get_array_xcorr(new_value)
        return self

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.revert()

    def __repr__(self):
        """ this hides the fact _Context instance are returned after calls """
        name = self.cache[self.value_to_switch].__str__()
        if hasattr(self.cache[self.value_to_switch], '__name__'):
            name = self.cache[self.value_to_switch].__name__
        out_str = ("%s changed to %s" % (self.value_to_switch, name))
        return out_str

    def revert(self):
        """ revert the default xcorr function to previous value """
        # Have to use the previous value as this may contain some custom
        # stream_xcorr functions
        self.cache[self.value_to_switch] = self.previous_value


set_xcorr = _Context(XCOR_FUNCS, 'default')


# ---------------------- generic concurrency functions

@contextlib.contextmanager
def pool_boy(Pool, traces, **kwargs):
    """
    A context manager for handling the setup and cleanup of a pool object.

    :param Pool: any Class (not instance) that implements the multiprocessing
        Pool interface
    :param traces: The number of traces to process
    :type traces: int
    """
    # All parallel processing happens on a per-trace basis, we shouldn't create
    # more workers than there are traces
    n_cores = kwargs.get('cores', cpu_count())
    if n_cores is None:
        n_cores = cpu_count()
    if n_cores > traces:
        n_cores = traces
    pool = Pool(n_cores)
    yield pool
    pool.close()
    pool.join()


def _pool_normxcorr(templates, stream, stack, pool, func, *args, **kwargs):
    chans = [[] for _i in range(len(templates))]
    array_dict_tuple = _get_array_dicts(templates, stream, stack=stack)
    stream_dict, template_dict, pad_dict, seed_ids = array_dict_tuple
    # get parameter iterator
    params = ((template_dict[sid], stream_dict[sid], pad_dict[sid])
              for sid in seed_ids)
    # get cc results and used chans into their own lists
    results = [pool.apply_async(func, param) for param in params]
    try:
        xcorrs, tr_chans = zip(*(res.get() for res in results))
    except KeyboardInterrupt as e:  # pragma: no cover
        pool.terminate()
        raise e
    if stack:
        cccsums = np.sum(xcorrs, axis=0)
    else:
        cccsums = np.asarray(xcorrs).swapaxes(0, 1)
    no_chans = np.sum(np.array(tr_chans).astype(int), axis=0)
    for seed_id, tr_chan in zip(seed_ids, tr_chans):
        for chan, state in zip(chans, tr_chan):
            if state:
                chan.append(seed_id)
    if stack:
        cccsums = _zero_invalid_correlation_sums(cccsums, pad_dict, chans)
    chans = [[(seed_id.split('.')[1], seed_id.split('.')[-1].split('_')[0])
              for seed_id in _chans] for _chans in chans]
    return cccsums, no_chans, chans


def _general_multithread(func):
    """ return the general multithreading function using func """

    def multithread(templates, stream, stack=True, *args, **kwargs):
        with pool_boy(ThreadPool, len(stream), **kwargs) as pool:
            return _pool_normxcorr(
                templates, stream, stack=stack, pool=pool, func=func)

    return multithread


def _general_multiprocess(func):
    def multiproc(templates, stream, stack=True, *args, **kwargs):
        with pool_boy(ProcessPool, len(stream), **kwargs) as pool:
            return _pool_normxcorr(
                templates, stream, stack=stack, pool=pool, func=func)

    return multiproc


def _general_serial(func):
    def stream_xcorr(templates, stream, stack=True, *args, **kwargs):
        no_chans = np.zeros(len(templates))
        chans = [[] for _ in range(len(templates))]
        array_dict_tuple = _get_array_dicts(templates, stream, stack=stack)
        stream_dict, template_dict, pad_dict, seed_ids = array_dict_tuple
        if stack:
            cccsums = np.zeros([len(templates),
                                len(stream[0]) - len(templates[0][0]) + 1])
        else:
            cccsums = np.zeros([len(templates), len(seed_ids),
                                len(stream[0]) - len(templates[0][0]) + 1])
        for chan_no, seed_id in enumerate(seed_ids):
            tr_cc, tr_chans = func(template_dict[seed_id],
                                   stream_dict[seed_id],
                                   pad_dict[seed_id])
            if stack:
                cccsums = np.sum([cccsums, tr_cc], axis=0)
            else:
                cccsums[:, chan_no] = tr_cc
            no_chans += tr_chans.astype(int)
            for chan, state in zip(chans, tr_chans):
                if state:
                    chan.append(seed_id)
        if stack:
            cccsums = _zero_invalid_correlation_sums(cccsums, pad_dict, chans)
        chans = [[(seed_id.split('.')[1], seed_id.split('.')[-1].split('_')[0])
                  for seed_id in _chans] for _chans in chans]
        return cccsums, no_chans, chans

    return stream_xcorr


def register_array_xcorr(name, func=None, is_default=False):
    """
    Decorator for registering correlation functions.

    Each function must have the same interface as numpy_normxcorr, which is
    *f(templates, stream, pads, *args, **kwargs)* any number of specific kwargs
    can be used.

    Register_normxcorr can be used as a decorator (with or without arguments)
    or as a callable.

    :param name: The name of the function for quick access, or the callable
        that will be wrapped when used as a decorator.
    :type name: str, callable
    :param func: The function to register
    :type func: callable, optional
    :param is_default: True if this function should be marked as default
        normxcorr
    :type is_default: bool

    :return: callable
    """
    valid_methods = set(list(XCOR_ARRAY_METHODS) + list(XCORR_STREAM_METHODS))
    cache = {}

    def register(register_str):
        """
        Register a function as an implementation.

        :param register_str: The registration designation
        :type register_str: str
        """
        if register_str not in valid_methods:
            msg = 'register_name must be in %s' % valid_methods
            raise ValueError(msg)

        def _register(func):
            cache[register_str] = func
            setattr(cache['func'], register_str, func)
            return func

        return _register

    def wrapper(func, func_name=None):
        # register the functions in the XCOR
        fname = func_name or name.__name__ if callable(name) else str(name)
        XCOR_FUNCS[fname] = func
        # if is_default:  # set function as default
        #     XCOR_FUNCS['default'] = func
        # attach some attrs, this is a bit of a hack to avoid pickle problems
        func.register = register
        cache['func'] = func
        func.multithread = _general_multithread(func)
        func.multiprocess = _general_multiprocess(func)
        func.concurrent = _general_multithread(func)
        func.stream_xcorr = _general_serial(func)
        func.array_xcorr = func
        func.registered = True
        if is_default:  # set function as default
            XCOR_FUNCS['default'] = copy.deepcopy(func)
        return func

    # used as a decorator
    if callable(name):
        return wrapper(name)

    # used as a normal function (called and passed a function)
    if callable(func):
        return wrapper(func, func_name=name)

    # called, then used as a decorator
    return wrapper


def _zero_invalid_correlation_sums(cccsums, pad_dict, used_seed_ids):
    """
    Zero the end portion of the correlation sum that does not include the full
    set of channels.

    :param cccsums: 2D numpy array
    :type cccsums: np.ndarray
    :param pad_dict: pad_dict from _get_array_dicts
    :type pad_dict: dict
    :param used_seed_ids: The SEED IDs actually used in correlation
    :type used_seed_ids: list of list of str

    :return:
        Valid correlation stack - end will be zero-ed up to maxmimum moveout
    :rtype: np.ndarray
    """
    # TODO: This is potentially quite a slow way to do this.
    for i, cccsum in enumerate(cccsums):
        moveouts = [value[i] for key, value in pad_dict.items()
                    if key in used_seed_ids[i]]
        max_moveout = 0
        if len(moveouts):
            max_moveout = max(moveouts)
        if max_moveout:
            cccsum[-max_moveout:] = 0.0
    return cccsums

# ------------------ array_xcorr fetching functions


def _get_registerd_func(name_or_func):
    """ get a xcorr function from a str or callable. """
    # get the function or register callable
    if callable(name_or_func):
        func = register_array_xcorr(name_or_func)
    else:
        func = XCOR_FUNCS[name_or_func or 'default']
    assert callable(func), 'func is not callable'
    # ensure func has the added methods
    if not hasattr(func, 'registered'):
        func = register_array_xcorr(func)
    return func


def get_array_xcorr(name_or_func=None):
    """
    Get an normalized cross correlation function that takes arrays as inputs.

    See :func:`eqcorrscan.utils.correlate.array_normxcorr` for expected
    function signature.

    :param name_or_func: Either a name of a registered xcorr function or a
        callable that implements the standard array_normxcorr signature.
    :type name_or_func: str or callable

    :return: callable wth array_normxcorr interface

    see also :func:`eqcorrscan.utils.correlate.get_stream_xcorr`
    """
    func = _get_registerd_func(name_or_func)
    return func.array_xcorr


# ----------------------- registered array_xcorr functions


@register_array_xcorr('numpy')
def numpy_normxcorr(templates, stream, pads, *args, **kwargs):
    """
    Compute the normalized cross-correlation using numpy and bottleneck.

    :param templates: 2D Array of templates
    :type templates: np.ndarray
    :param stream: 1D array of continuous data
    :type stream: np.ndarray
    :param pads: List of ints of pad lengths in the same order as templates
    :type pads: list

    :return: np.ndarray of cross-correlations
    :return: np.ndarray channels used
    """
    import bottleneck

    # Generate a template mask
    used_chans = ~np.isnan(templates).any(axis=1)
    # Currently have to use float64 as bottleneck runs into issues with other
    # types: https://github.com/kwgoodman/bottleneck/issues/164
    stream = stream.astype(np.float64)
    templates = templates.astype(np.float64)
    template_length = templates.shape[1]
    stream_length = len(stream)
    assert stream_length > template_length, "Template must be shorter than " \
                                            "stream"
    fftshape = next_fast_len(template_length + stream_length - 1)
    # Set up normalizers
    stream_mean_array = bottleneck.move_mean(
        stream, template_length)[template_length - 1:]
    stream_std_array = bottleneck.move_std(
        stream, template_length)[template_length - 1:]
    # because stream_std_array is in denominator or res, nan all 0s
    stream_std_array[stream_std_array == 0] = np.nan
    # Normalize and flip the templates
    norm = ((templates - templates.mean(axis=-1, keepdims=True)) / (
        templates.std(axis=-1, keepdims=True) * template_length))
    norm_sum = norm.sum(axis=-1, keepdims=True)
    stream_fft = np.fft.rfft(stream, fftshape)
    template_fft = np.fft.rfft(np.flip(norm, axis=-1), fftshape, axis=-1)
    res = np.fft.irfft(template_fft * stream_fft,
                       fftshape)[:, 0:template_length + stream_length - 1]
    res = ((_centered(res, (templates.shape[0],
                            stream_length - template_length + 1))) -
           norm_sum * stream_mean_array) / stream_std_array
    res[np.isnan(res)] = 0.0

    for i, pad in enumerate(pads):
        res[i] = np.append(res[i], np.zeros(pad))[pad:]
    return res.astype(np.float32), used_chans


def _centered(arr, newshape):
    """
    Hack of scipy.signaltools._centered
    """
    newshape = np.asarray(newshape)
    currshape = np.array(arr.shape)
    startind = (currshape - newshape) // 2
    endind = startind + newshape
    myslice = [slice(startind[k], endind[k]) for k in range(len(endind))]
    return arr[tuple(myslice)]


@register_array_xcorr('time_domain')
def time_multi_normxcorr(templates, stream, pads, threaded=False, *args,
                         **kwargs):
    """
    Compute cross-correlations in the time-domain using C routine.

    :param templates: 2D Array of templates
    :type templates: np.ndarray
    :param stream: 1D array of continuous data
    :type stream: np.ndarray
    :param pads: List of ints of pad lengths in the same order as templates
    :type pads: list
    :param threaded: Whether to use the threaded routine or not
    :type threaded: bool

    :return: np.ndarray of cross-correlations
    :return: np.ndarray channels used
    """
    used_chans = ~np.isnan(templates).any(axis=1)

    utilslib = _load_cdll('libutils')

    argtypes = [
        np.ctypeslib.ndpointer(dtype=np.float32, ndim=1,
                               flags='C_CONTIGUOUS'),
        ctypes.c_int, ctypes.c_int,
        np.ctypeslib.ndpointer(dtype=np.float32, ndim=1,
                               flags='C_CONTIGUOUS'),
        ctypes.c_int,
        np.ctypeslib.ndpointer(dtype=np.float32, ndim=1,
                               flags='C_CONTIGUOUS')]
    restype = ctypes.c_int
    if threaded:
        func = utilslib.multi_normxcorr_time_threaded
        argtypes.append(ctypes.c_int)
    else:
        func = utilslib.multi_normxcorr_time
    func.argtypes = argtypes
    func.restype = restype
    # Need to de-mean everything
    templates_means = templates.mean(axis=1).astype(np.float32)[:, np.newaxis]
    stream_mean = stream.mean().astype(np.float32)
    templates = templates.astype(np.float32) - templates_means
    stream = stream.astype(np.float32) - stream_mean
    template_len = templates.shape[1]
    n_templates = templates.shape[0]
    image_len = stream.shape[0]
    ccc_length = image_len - template_len + 1
    assert ccc_length > 0, "Template must be shorter than stream"
    ccc = np.ascontiguousarray(
        np.empty(ccc_length * n_templates), np.float32)
    t_array = np.ascontiguousarray(templates.flatten(), np.float32)
    time_args = [t_array, template_len, n_templates,
                 np.ascontiguousarray(stream, np.float32), image_len, ccc]
    if threaded:
        time_args.append(kwargs.get('cores', cpu_count()))
    func(*time_args)
    ccc[np.isnan(ccc)] = 0.0
    ccc = ccc.reshape((n_templates, image_len - template_len + 1))
    for i in range(len(pads)):
        ccc[i] = np.append(ccc[i], np.zeros(pads[i]))[pads[i]:]
    templates += templates_means
    stream += stream_mean
    return ccc, used_chans


@register_array_xcorr('fftw', is_default=True)
def fftw_normxcorr(templates, stream, pads, threaded=False, *args, **kwargs):
    """
    Normalised cross-correlation using the fftw library.

    Internally this function used double precision numbers, which is definitely
    required for seismic data. Cross-correlations are computed as the
    inverse fft of the dot product of the ffts of the stream and the reversed,
    normalised, templates.  The cross-correlation is then normalised using the
    running mean and standard deviation (not using the N-1 correction) of the
    stream and the sums of the normalised templates.

    This python function wraps the C-library written by C. Chamberlain for this
    purpose.

    :param templates: 2D Array of templates
    :type templates: np.ndarray
    :param stream: 1D array of continuous data
    :type stream: np.ndarray
    :param pads: List of ints of pad lengths in the same order as templates
    :type pads: list
    :param threaded:
        Whether to use the threaded routine or not - note openMP and python
        multiprocessing don't seem to play nice for this.
    :type threaded: bool

    :return: np.ndarray of cross-correlations
    :return: np.ndarray channels used
    """
    utilslib = _load_cdll('libutils')

    argtypes = [
        np.ctypeslib.ndpointer(dtype=np.float32, ndim=1,
                               flags='C_CONTIGUOUS'),
        ctypes.c_long, ctypes.c_long,
        np.ctypeslib.ndpointer(dtype=np.float32, ndim=1,
                               flags='C_CONTIGUOUS'),
        ctypes.c_long,
        np.ctypeslib.ndpointer(dtype=np.float32,
                               flags='C_CONTIGUOUS'),
        ctypes.c_long,
        np.ctypeslib.ndpointer(dtype=np.intc,
                               flags='C_CONTIGUOUS'),
        np.ctypeslib.ndpointer(dtype=np.intc,
                               flags='C_CONTIGUOUS'),
        np.ctypeslib.ndpointer(dtype=np.intc,
                               flags='C_CONTIGUOUS'),
        np.ctypeslib.ndpointer(dtype=np.intc,
                               flags='C_CONTIGUOUS')]
    restype = ctypes.c_int

    if threaded:
        func = utilslib.normxcorr_fftw_threaded
    else:
        func = utilslib.normxcorr_fftw

    func.argtypes = argtypes
    func.restype = restype

    # Generate a template mask
    used_chans = ~np.isnan(templates).any(axis=1)
    template_length = templates.shape[1]
    stream_length = len(stream)
    n_templates = templates.shape[0]
    fftshape = kwargs.get("fft_len")
    if fftshape is None:
        # In testing, 2**13 consistently comes out fastest - setting to
        # default. https://github.com/eqcorrscan/EQcorrscan/pull/285
        fftshape = min(
            2 ** 13, next_fast_len(template_length + stream_length - 1))
    if fftshape < template_length:
        Logger.warning(
            "FFT length of {0} is shorter than the template, setting to "
            "{1}".format(
                fftshape, next_fast_len(template_length + stream_length - 1)))
        fftshape = next_fast_len(template_length + stream_length - 1)
    # Normalize and flip the templates
    norm = ((templates - templates.mean(axis=-1, keepdims=True)) / (
        templates.std(axis=-1, keepdims=True) * template_length))

    norm = np.nan_to_num(norm)
    ccc_length = stream_length - template_length + 1
    assert ccc_length > 0, "Template must be shorter than stream"
    ccc = np.zeros((n_templates, ccc_length), np.float32)
    used_chans_np = np.ascontiguousarray(used_chans, dtype=np.intc)
    pads_np = np.ascontiguousarray(pads, dtype=np.intc)
    variance_warning = np.ascontiguousarray([0], dtype=np.intc)
    missed_corr = np.ascontiguousarray([0], dtype=np.intc)

    # Check that stream is non-zero and above variance threshold
    if not np.all(stream == 0) and np.var(stream) < 1e-8:
        # Apply gain
        stream *= MULTIPLIER
        Logger.warning("Low variance found for, applying gain "
                       "to stabilise correlations")
        multiplier = MULTIPLIER
    else:
        multiplier = 1
    ret = func(
        np.ascontiguousarray(norm.flatten(order='C'), np.float32),
        template_length, n_templates,
        np.ascontiguousarray(stream, np.float32), stream_length,
        np.ascontiguousarray(ccc, np.float32), fftshape,
        used_chans_np, pads_np, variance_warning, missed_corr)
    if ret < 0:
        raise MemoryError()
    elif ret > 0:
        Logger.critical('Error in C code (possible normalisation error)')
        Logger.critical('Maximum ccc %f at %i' % (ccc.max(), ccc.argmax()))
        Logger.critical('Minimum ccc %f at %i' % (ccc.min(), ccc.argmin()))
        Logger.critical('Recommend checking your data for spikes, clipping '
                        'or artefacts')
        raise CorrelationError("Internal correlation error")
    if missed_corr[0]:
        Logger.warning(
            "{0} correlations not computed, are there gaps in the "
            "data? If not, consider increasing gain".format(
                missed_corr[0]))
    if variance_warning[0] and variance_warning[0] > template_length:
        Logger.warning(
            "Low variance found in {0} positions, check result.".format(
                variance_warning[0]))
    # Remove variance correction
    stream /= multiplier
    return ccc, used_chans


# The time-domain routine can be sped up massively on large machines (many
# threads) using the openMP threaded functions.

@time_multi_normxcorr.register('concurrent')
def _time_threaded_normxcorr(templates, stream, stack=True, *args, **kwargs):
    """
    Use the threaded time-domain routine for concurrency

    :type templates: list
    :param templates:
        A list of templates, where each one should be an obspy.Stream object
        containing multiple traces of seismic data and the relevant header
        information.
    :type stream: obspy.core.stream.Stream
    :param stream:
        A single Stream object to be correlated with the templates.

    :returns:
        New list of :class:`numpy.ndarray` objects.  These will contain
        the correlation sums for each template for this day of data.
    :rtype: list
    :returns:
        list of ints as number of channels used for each cross-correlation.
    :rtype: list
    :returns:
        list of list of tuples of station, channel for all cross-correlations.
    :rtype: list
    """
    no_chans = np.zeros(len(templates))
    chans = [[] for _ in range(len(templates))]
    array_dict_tuple = _get_array_dicts(templates, stream, stack=stack)
    stream_dict, template_dict, pad_dict, seed_ids = array_dict_tuple
    ccc_length = max(
        len(stream[0]) - len(templates[0][0]) + 1,
        len(templates[0][0]) - len(stream[0]) + 1)
    if stack:
        cccsums = np.zeros([len(templates), ccc_length])
    else:
        cccsums = np.zeros([len(templates), len(seed_ids), ccc_length])
    for chan_no, seed_id in enumerate(seed_ids):
        tr_cc, tr_chans = time_multi_normxcorr(
            template_dict[seed_id], stream_dict[seed_id], pad_dict[seed_id],
            True)
        if stack:
            cccsums = np.sum([cccsums, tr_cc], axis=0)
        else:
            cccsums[:, chan_no] = tr_cc
        no_chans += tr_chans.astype(int)
        for chan, state in zip(chans, tr_chans):
            if state:
                chan.append(seed_id)
    if stack:
        cccsums = _zero_invalid_correlation_sums(cccsums, pad_dict, chans)
    chans = [[(seed_id.split('.')[1], seed_id.split('.')[-1].split('_')[0])
              for seed_id in _chans] for _chans in chans]
    return cccsums, no_chans, chans


@fftw_normxcorr.register('stream_xcorr')
@fftw_normxcorr.register('multithread')
@fftw_normxcorr.register('concurrent')
def _fftw_stream_xcorr(templates, stream, stack=True, *args, **kwargs):
    """
    Apply fftw normxcorr routine concurrently.

    :type templates: list
    :param templates:
        A list of templates, where each one should be an obspy.Stream object
        containing multiple traces of seismic data and the relevant header
        information.
    :type stream: obspy.core.stream.Stream
    :param stream:
        A single Stream object to be correlated with the templates.

    :returns:
        New list of :class:`numpy.ndarray` objects.  These will contain
        the correlation sums for each template for this day of data.
    :rtype: list
    :returns:
        list of ints as number of channels used for each cross-correlation.
    :rtype: list
    :returns:
        list of list of tuples of station, channel for all cross-correlations.
    :rtype: list
    """
    array_dict_tuple = _get_array_dicts(templates, stream, stack=stack)
    stream_dict, template_dict, pad_dict, seed_ids = array_dict_tuple
    assert set(seed_ids)
    # number of threads:
    #   default to using inner threads
    #   if `cores` or `cores_outer` passed in then use that
    #   else if OMP_NUM_THREADS set use that
    #   otherwise use all available
    inner_kwargs = copy.copy(kwargs)
    num_cores_inner, num_cores_outer = _set_inner_outer_threading(
        inner_kwargs.pop('cores', None),
        inner_kwargs.pop("cores_outer", None),
        len(stream))

    chans = [[] for _i in range(len(templates))]
    cccsums, tr_chans = fftw_multi_normxcorr(
        template_array=template_dict, stream_array=stream_dict,
        pad_array=pad_dict, seed_ids=seed_ids, cores_inner=num_cores_inner,
        cores_outer=num_cores_outer, stack=stack, *args, **inner_kwargs)
    no_chans = np.sum(np.array(tr_chans).astype(int), axis=0)
    for seed_id, tr_chan in zip(seed_ids, tr_chans):
        for chan, state in zip(chans, tr_chan):
            if state:
                chan.append(seed_id)
    if stack:
        cccsums = _zero_invalid_correlation_sums(cccsums, pad_dict, chans)
    chans = [[(seed_id.split('.')[1], seed_id.split('.')[-1].split('_')[0])
              for seed_id in _chans] for _chans in chans]
    return cccsums, no_chans, chans


def _set_inner_outer_threading(num_cores_inner, num_cores_outer, n_chans):
    max_threads = int(os.getenv("OMP_NUM_THREADS", cpu_count()))
    if num_cores_inner is None:
        num_cores_inner = max_threads
    num_cores_outer = num_cores_outer or 1
    if num_cores_outer > 1:
        if num_cores_outer > n_chans:
            Logger.info(
                "More outer cores than channels, setting to {0}".format(
                    n_chans))
            num_cores_outer = n_chans
        if num_cores_outer * num_cores_inner > max_threads:
            Logger.info("More threads requested than exist, falling back to "
                        "outer-loop parallelism")
            num_cores_outer = min(max_threads, num_cores_outer)
            if 2 * num_cores_outer < max_threads:
                num_cores_inner = max_threads // num_cores_outer
            else:
                num_cores_inner = 1
    return num_cores_inner, num_cores_outer


def fftw_multi_normxcorr(
    template_array,
    stream_array,
    pad_array,
    seed_ids,
    cores_inner,
    cores_outer,
    stack=True,
    *args,
    **kwargs
):
    """
    Use a C loop rather than a Python loop - in some cases this will be fast.

    :type template_array: dict
    :param template_array:
    :type stream_array: dict
    :param stream_array:
    :type pad_array: dict
    :param pad_array:
    :type seed_ids: list
    :param seed_ids:

    rtype: np.ndarray, list
    :return: 3D Array of cross-correlations and list of used channels.
    """
    utilslib = _load_cdll('libutils')

    utilslib.multi_normxcorr_fftw.argtypes = [
        np.ctypeslib.ndpointer(dtype=np.float32,
                               flags='C_CONTIGUOUS'),
        ctypes.c_long, ctypes.c_long, ctypes.c_long,
        np.ctypeslib.ndpointer(dtype=np.float32,
                               flags='C_CONTIGUOUS'),
        ctypes.c_long,
        np.ctypeslib.ndpointer(dtype=np.float32,
                               flags='C_CONTIGUOUS'),
        ctypes.c_long,
        np.ctypeslib.ndpointer(dtype=np.intc,
                               flags='C_CONTIGUOUS'),
        np.ctypeslib.ndpointer(dtype=np.intc,
                               flags='C_CONTIGUOUS'),
        ctypes.c_int, ctypes.c_int,
        np.ctypeslib.ndpointer(dtype=np.intc,
                               flags='C_CONTIGUOUS'),
        np.ctypeslib.ndpointer(dtype=np.intc,
                               flags='C_CONTIGUOUS'),
        ctypes.c_int]
    utilslib.multi_normxcorr_fftw.restype = ctypes.c_int
    '''
    Arguments are:
        templates (stacked [ch_1-t_1, ch_1-t_2, ..., ch_2-t_1, ch_2-t_2, ...])
        number of templates
        template length
        number of channels
        image (stacked [ch_1, ch_2, ..., ch_n])
        image length
        cross-correlations (stacked as per image)
        fft-length
        used channels (stacked as per templates)
        pad array (stacked as per templates)
        num threads inner
        num threads outer
        variance warnings
        missed correlation warnings (usually due to gaps)
        stack option
    '''

    # pre processing
    used_chans = []
    template_len = template_array[seed_ids[0]].shape[1]
    for seed_id in seed_ids:
        used_chans.append(~np.isnan(template_array[seed_id]).any(axis=1))
        template_array[seed_id] = (
            (template_array[seed_id] -
             template_array[seed_id].mean(axis=-1, keepdims=True)) / (
                template_array[seed_id].std(axis=-1, keepdims=True) *
                template_len))
        template_array[seed_id] = np.nan_to_num(template_array[seed_id])
    n_channels = len(seed_ids)
    n_templates = template_array[seed_ids[0]].shape[0]
    image_len = stream_array[seed_ids[0]].shape[0]
    fft_len = kwargs.get("fft_len")
    if fft_len is None:
        # In testing, 2**13 consistently comes out fastest - setting to
        # default. https://github.com/eqcorrscan/EQcorrscan/pull/285
        # But this results in lots of chunks - 2 ** 17 is also good.
        fft_len = min(2 ** 17, next_fast_len(template_len + image_len - 1))
    if fft_len < template_len:
        Logger.warning(
            f"FFT length of {fft_len} is shorter than the template, setting to"
            f" {next_fast_len(template_len + image_len - 1)}")
        fft_len = next_fast_len(template_len + image_len - 1)
    template_array = np.ascontiguousarray(
        [template_array[x] for x in seed_ids], dtype=np.float32)
    multipliers = {}
    for x in seed_ids:
        # Check that stream is non-zero and above variance threshold
        if not np.all(stream_array[x] == 0) and np.var(stream_array[x]) < 1e-8:
            # Apply gain
            stream_array[x] *= MULTIPLIER
            Logger.warning(f"Low variance found for {x}, applying gain "
                           "to stabilise correlations")
            multipliers.update({x: MULTIPLIER})
        else:
            multipliers.update({x: 1})
    stream_array = np.ascontiguousarray([stream_array[x] for x in seed_ids],
                                        dtype=np.float32)
    ccc_length = image_len - template_len + 1
    assert ccc_length > 0, "Template must be shorter than stream"
    if stack:
        cccs = np.zeros((n_templates, ccc_length), np.float32)
    else:
        cccs = np.zeros(
            (n_templates, n_channels, ccc_length), dtype=np.float32)
    used_chans_np = np.ascontiguousarray(used_chans, dtype=np.intc)
    pad_array_np = np.ascontiguousarray(
        [pad_array[seed_id] for seed_id in seed_ids], dtype=np.intc)
    variance_warnings = np.ascontiguousarray(
        np.zeros(n_channels), dtype=np.intc)
    missed_correlations = np.ascontiguousarray(
        np.zeros(n_channels), dtype=np.intc)

    # call C function
    ret = utilslib.multi_normxcorr_fftw(
        template_array, n_templates, template_len, n_channels, stream_array,
        image_len, cccs, fft_len, used_chans_np, pad_array_np,
        cores_inner, cores_outer, variance_warnings, missed_correlations,
        int(stack))
    if ret < 0:
        raise MemoryError("Memory allocation failed in correlation C-code")
    elif ret > 0:
        Logger.critical(
            'Out-of-range correlation in C-code, see WARNING from C-code.'
            'You are STRONGLY RECOMMENDED to check your data for spikes, '
            'clipping or non-physical artifacts')
        # raise CorrelationError("Internal correlation error")
    for i, missed_corr in enumerate(missed_correlations):
        if missed_corr:
            Logger.debug(
                f"{missed_corr} correlations not computed on {seed_ids[i]}, "
                f"are there gaps in the data? If not, consider "
                "increasing gain")
    for i, variance_warning in enumerate(variance_warnings):
        if variance_warning and variance_warning > template_len:
            Logger.warning(
                f"Low variance found in {variance_warning} places for "
                f"{seed_ids[i]}, check result.")
    # Remove gain
    for i, x in enumerate(seed_ids):
        stream_array[i] *= multipliers[x]
    return cccs, used_chans


# ------------------------------- FastMatchedFilter Wrapper

def _run_fmf_xcorr(template_arr, data_arr, weights, pads, arch, step=1):
    template_arr, data_arr, multipliers = _fmf_stabilisation(
        template_arr=template_arr, data_arr=data_arr)

    return _stabalised_fmf(
        template_arr=template_arr, data_arr=data_arr, weights=weights,
        pads=pads, arch=arch, multipliers=multipliers, step=step)


def _stabalised_fmf(template_arr, data_arr, weights, pads, arch, multipliers,
                    step):
    if not FMF_INSTALLED:
        raise ImportError("FastMatchedFilter is not available")
    import fast_matched_filter

    if version.parse(fast_matched_filter.__version__) >= MIN_FMF_VERSION:
        from fast_matched_filter import matched_filter as fmf
    else:
        raise ImportError(f"FMF version {fast_matched_filter.__version__} "
                          f"must be >= {MIN_FMF_VERSION}")

    Logger.info("Handing off to FMF")
    cccsums = fmf(
        templates=template_arr, weights=weights, moveouts=pads,
        data=data_arr, step=step, arch=arch, normalize="full")
    Logger.info("FMF returned")
    # Remove gain
    if np.any(multipliers != 1):
        data_arr /= multipliers
    return cccsums


def _fmf_stabilisation(template_arr, data_arr):
    """ FMF doesn't do the internal check that EQC correlations do. """
    # Demean
    tic = default_timer()
    template_arr -= template_arr.mean(axis=-1, keepdims=True)
    data_arr -= data_arr.mean(axis=-1, keepdims=True)
    toc = default_timer()
    Logger.info(f"Removing mean took {toc - tic:.4f} s")

    # Stability checking
    tic = default_timer()
    # var is fairly slow, var = mean(abs(a - a.mean()) ** 2) - mean is zero,
    # so we can skip a step
    stability_issues = np.logical_and(
        # data_arr.var(axis=1, keepdims=True) < 1e-8,
        np.mean(np.abs(data_arr) ** 2, axis=1, keepdims=True) < 1e-8,
        ~np.all(data_arr == 0, axis=1, keepdims=True))
    multipliers = np.ones_like(stability_issues, dtype=float)
    multipliers[stability_issues] = MULTIPLIER
    if np.any(stability_issues):
        Logger.warning(
            f"Low variance found for channels {np.where(stability_issues)},"
            f"applying gain to stabilise correlations")
        data_arr *= multipliers
    toc = default_timer()
    Logger.info(f"Checking stability took {toc - tic:.4f} s")
    return template_arr, data_arr, multipliers


@register_array_xcorr("fmf")
def fmf_xcorr(templates, stream, pads, arch="precise", *args, **kwargs):
    """
    Compute cross-correlations in the time-domain using the FMF routine.

    :param templates: 2D Array of templates
    :type templates: np.ndarray
    :param stream: 1D array of continuous data
    :type stream: np.ndarray
    :param pads: List of ints of pad lengths in the same order as templates
    :type pads: list
    :param arch: "gpu" or "precise" to run on GPU or CPU respectively
    type arch: str

    :return: np.ndarray of cross-correlations
    :return: np.ndarray channels used
    """
    assert templates.ndim == 2, "Templates must be 2D"
    assert stream.ndim == 1, "Stream must be 1D"

    used_chans = ~np.isnan(templates).any(axis=1)

    # We have to reshape to an extra dimension for FMF
    ccc = _run_fmf_xcorr(
        template_arr=templates.reshape(
            (1, templates.shape[0], templates.shape[1])).swapaxes(0, 1),
        data_arr=stream.reshape((1, stream.shape[0])),
        weights=np.ones((1, templates.shape[0])),
        pads=np.array([pads]),
        arch=arch.lower())

    return ccc, used_chans


@fmf_xcorr.register("stream_xcorr")
@fmf_xcorr.register("concurrent")
def _fmf_gpu(templates, stream, *args, **kwargs):
    """
    Thin wrapper of fmf_multi_xcorr setting arch to gpu.
    """
    from fast_matched_filter import GPU_LOADED
    if not GPU_LOADED:
        Logger.warning("FMF reports GPU not loaded, reverting to CPU")
        return _fmf_cpu(templates=templates, stream=stream, *args, **kwargs)
    return _fmf_multi_xcorr(templates, stream, arch="gpu")


@fmf_xcorr.register("multithread")
@fmf_xcorr.register("multiprocess")
def _fmf_cpu(templates, stream, *args, **kwargs):
    """
    Thin wrapper of fmf_multi_xcorr setting arch to cpu.
    """
    from fast_matched_filter import CPU_LOADED
    if not CPU_LOADED:
        raise NotImplementedError(
            "FMF reports CPU not loaded - try rebuilding FMF")
    return _fmf_multi_xcorr(templates, stream, arch="precise")


def _fmf_reshape(template_dict, stream_dict, pad_dict, seed_ids):
    tic = default_timer()
    # Reshape templates into [templates x traces x time]
    t_arr = np.array([template_dict[seed_id]
                      for seed_id in seed_ids]).swapaxes(0, 1)
    # Reshape stream into [traces x time]
    d_arr = np.array([stream_dict[seed_id] for seed_id in seed_ids])
    # Moveouts should be [templates x traces]
    pads = np.array([pad_dict[seed_id] for seed_id in seed_ids]).swapaxes(0, 1)
    # Weights should be shaped like pads
    weights = np.ones_like(pads)
    toc = default_timer()
    Logger.info(f"Reshaping for FMF took {toc - tic:.4f} s")
    return t_arr, d_arr, weights, pads


def _fmf_multi_xcorr(templates, stream, *args, **kwargs):
    """
    Apply FastMatchedFilter routine concurrently.

    :type templates: list
    :param templates:
        A list of templates, where each one should be an obspy.Stream object
        containing multiple traces of seismic data and the relevant header
        information.
    :type stream: obspy.core.stream.Stream
    :param stream:
        A single Stream object to be correlated with the templates.

    :returns:
        New list of :class:`numpy.ndarray` objects.  These will contain
        the correlation sums for each template for this day of data.
    :rtype: list
    :returns:
        list of ints as number of channels used for each cross-correlation.
    :rtype: list
    :returns:
        list of list of tuples of station, channel for all cross-correlations.
    :rtype: list
    """
    if kwargs.get("stack", False):
        raise NotImplementedError(
            "FMF does not support unstacked correlations, use a different "
            "backend")
    arch = kwargs.get("arch", "gpu")
    Logger.info(f"Running FMF targeting the {arch}")

    chans = [[] for _i in range(len(templates))]
    array_dict_tuple = _get_array_dicts(templates, stream, stack=True)
    stream_dict, template_dict, pad_dict, seed_ids = array_dict_tuple
    assert set(seed_ids)

    t_arr, d_arr, weights, pads = _fmf_reshape(
        template_dict=template_dict, stream_dict=stream_dict,
        pad_dict=pad_dict, seed_ids=seed_ids)

    cccsums = _run_fmf_xcorr(
        template_arr=t_arr, weights=weights, pads=pads,
        data_arr=d_arr, step=1, arch=arch)

    tr_chans = np.array([~np.isnan(template_dict[seed_id]).any(axis=1)
                         for seed_id in seed_ids])
    no_chans = np.sum(np.array(tr_chans).astype(int), axis=0)
    # Note: FMF already returns the zeroed end of correlations - we don't
    # need to call _get_valid_correlation_sum
    for seed_id, tr_chan in zip(seed_ids, tr_chans):
        for chan, state in zip(chans, tr_chan):
            if state:
                chan.append((seed_id.split('.')[1],
                             seed_id.split('.')[-1].split('_')[0]))
    return cccsums, no_chans, chans


# ------------------------------- stream_xcorr functions


def get_stream_xcorr(name_or_func=None, concurrency=None):
    """
    Return a function for performing normalized cross correlation on lists of
    streams.

    :param name_or_func:
        Either a name of a registered function or a callable that implements
        the standard array_normxcorr signature.
    :param concurrency:
        Optional concurrency strategy, options are below.

    :return: A callable with the interface of stream_normxcorr

    :Concurrency options:
        - multithread - use a threadpool for concurrency;
        - multiprocess - use a process pool for concurrency;
        - concurrent - use a customized concurrency strategy for the function,
          if not defined threading will be used.
    """
    func = _get_registerd_func(name_or_func)

    concur = concurrency or 'stream_xcorr'
    if not hasattr(func, concur):
        msg = '%s does not support concurrency %s' % (func.__name__, concur)
        raise ValueError(msg)
    return getattr(func, concur)


# --------------------------- stream prep functions


def _get_array_dicts(templates, stream, stack, *args, **kwargs):
    """ prepare templates and stream, return dicts """
    tic = default_timer()
    # Do some reshaping
    # init empty structures for data storage
    template_dict = {}
    stream_dict = {}
    pad_dict = {}
    t_starts = []

    stream.sort(['network', 'station', 'location', 'channel'])
    for template in templates:
        template.sort(['network', 'station', 'location', 'channel'])
        t_starts.append(
            UTCDateTime(ns=min([tr.stats.starttime.__dict__['_UTCDateTime__ns']
                                for tr in template])))
    stream_start = min([tr.stats.starttime for tr in stream])
    # get seed ids, make sure these are collected on sorted streams
    seed_ids = [tr.id + '_' + str(i) for i, tr in enumerate(templates[0])]
    # pull common channels out of streams and templates and put in dicts
    for i, seed_id in enumerate(seed_ids):
        temps_with_seed = [template.traces[i].data for template in templates]
        t_ar = np.array(temps_with_seed).astype(np.float32)
        template_dict.update({seed_id: t_ar})
        stream_channel = _stream_quick_select(stream, seed_id.split('_')[0])[0]
        # Normalize data to ensure no float overflow
        stream_data = stream_channel.data / (np.max(
            np.abs(stream_channel.data)) / 1e5)
        stream_dict.update(
            {seed_id: stream_data.astype(np.float32)})
        # PROBLEM - DEBUG: if two traces start just before / just after a
        # "full-sample-time", then stream_offset can become 1, while a value in
        # pad_list can become 0. 0-1 = -1; which is problematic.
        stream_offset = int(
            math.floor(stream_channel.stats.sampling_rate *
                       (stream_channel.stats.starttime - stream_start)))
        if stack:
            pad_list = [
                int(round(
                    template.traces[i].stats.__dict__['sampling_rate'] *
                    (template.traces[i].stats.starttime.__dict__[
                        '_UTCDateTime__ns'] -
                     t_starts[j].__dict__['_UTCDateTime__ns']) / 1e9)) -
                stream_offset
                for j, template in enumerate(templates)]
        else:
            pad_list = [0 for _ in templates]
        pad_dict.update({seed_id: pad_list})
    toc = default_timer()
    Logger.info(f"Making array dicts for {len(seed_ids)} seed ids "
                f"took {toc - tic:.4f} s")

    return stream_dict, template_dict, pad_dict, seed_ids


# Remove fmf if it isn't installed
if not FMF_INSTALLED:
    XCOR_FUNCS.pop("fmf")
# a dict of built in xcorr functions, used to distinguish from user-defined
XCORR_FUNCS_ORIGINAL = copy.copy(XCOR_FUNCS)


if __name__ == '__main__':
    import doctest

    doctest.testmod()



"""
A series of test functions for the core functions in EQcorrscan.
"""
import glob
import os
import shutil
import unittest
import numpy as np
import logging

from obspy import read, UTCDateTime

from eqcorrscan.utils.synth_seis import generate_synth_data
from eqcorrscan.utils.correlate import get_stream_xcorr
from eqcorrscan.core import lag_calc as lag_calc_module
from eqcorrscan.core.lag_calc import (
    _xcorr_interp, LagCalcError, _prepare_data, lag_calc, xcorr_pick_family,
    _concatenate_and_correlate)
from eqcorrscan.core.match_filter import Detection, Family, Party, Template
from eqcorrscan.helpers.mock_logger import MockLoggingHandler


class SyntheticTests(unittest.TestCase):
    """ Test lag-calc with synthetic data. """
    @classmethod
    def setUpClass(cls):
        np.random.seed(999)
        print("Setting up class")
        np.random.seed(999)
        samp_rate = 50
        t_length = .75
        # Make some synthetic templates
        templates, data, seeds = generate_synth_data(
            nsta=5, ntemplates=5, nseeds=10, samp_rate=samp_rate,
            t_length=t_length, max_amp=10, max_lag=15, phaseout="both",
            jitter=0, noise=False, same_phase=True)
        # Rename channels
        channel_mapper = {"SYN_Z": "HHZ", "SYN_H": "HHN"}
        for tr in data:
            tr.stats.channel = channel_mapper[tr.stats.channel]
        for template in templates:
            for tr in template:
                tr.stats.channel = channel_mapper[tr.stats.channel]
        party = Party()
        t = 0
        data_start = data[0].stats.starttime
        for template, template_seeds in zip(templates, seeds):
            template_name = "template_{0}".format(t)
            detections = []
            for i, sample in enumerate(template_seeds["time"]):
                det = Detection(
                    template_name=template_name,
                    detect_time=data_start + (sample / samp_rate),
                    detect_val=template_seeds["SNR"][i] / len(data),
                    no_chans=len(data),
                    chans=[
                        (tr.stats.station, tr.stats.channel) for tr in data],
                    threshold=0.0, threshold_input=0.0, threshold_type="abs",
                    typeofdet="ccc")
                det._calculate_event(
                    template_st=template, estimate_origin=False)
                detections.append(det)
            # Make a fully formed Template
            _template = Template(
                name=template_name, st=template, lowcut=2.0, highcut=15.0,
                samp_rate=samp_rate, filt_order=4, process_length=86400,
                prepick=10. / samp_rate, event=None)
            family = Family(template=_template, detections=detections)
            party += family
            t += 1
        cls.party = party
        cls.data = data
        cls.t_length = t_length
        assert len(data) == 10

    def _prepare_data_checks(self, detect_stream_dict, shift_len, family):
        self.assertEqual(
            len(detect_stream_dict), len(family.detections))
        for detection_id, stream in detect_stream_dict.items():
            detection = [d for d in family if d.id == detection_id][0]
            self.assertEqual(len(stream), detection.no_chans)
            for tr in stream:
                self.assertAlmostEqual(
                    tr.stats.endtime - tr.stats.starttime,
                    (2 * shift_len) + self.t_length, 1)
                pick = [p for p in detection.event.picks
                        if p.waveform_id.get_seed_string() == tr.id][0]
                self.assertAlmostEqual(
                    tr.stats.starttime,
                    pick.time - (family.template.prepick + shift_len), 1)

    def test_prepare_data(self):
        shift_len = 0.2
        for family in self.party:
            detect_stream_dict = _prepare_data(
                family=family, detect_data=self.data, shift_len=shift_len)
            self._prepare_data_checks(detect_stream_dict=detect_stream_dict,
                                      family=family, shift_len=shift_len)

    def test_prepare_data_too_short(self):
        data = self.data.copy()
        data.trim(self.party[0][0].detect_time,
                  self.party[0][0].detect_time + 3)
        shift_len = 0.2
        detect_stream_dict = _prepare_data(
            family=self.party[0], detect_data=data, shift_len=shift_len)
        self.assertEqual(len(detect_stream_dict), 1)

    def test_prepare_data_masked(self):
        data = self.data.copy()
        data.cutout(self.party[0][0].detect_time,
                    self.party[0][0].detect_time + 3)
        data.merge()
        shift_len = 0.2
        detect_stream_dict = _prepare_data(
            family=self.party[0], detect_data=data, shift_len=shift_len)
        short_key = self.party[0][0].id
        for key, value in detect_stream_dict.items():
            detection = [d for d in self.party[0] if d.id == key][0]
            if key == short_key:
                self.assertNotEqual(len(value), detection.no_chans)
            else:
                self.assertEqual(len(value), detection.no_chans)

    def test_family_picking(self):
        catalog_dict = xcorr_pick_family(
            family=self.party[0], stream=self.data, shift_len=0.2, plot=False,
            export_cc=False)
        for event in catalog_dict.values():
            self.assertEqual(len(event.picks), len(self.data))
            for pick in event.picks:
                self.assertTrue("cc_max=" in pick.comments[0].text)
                self.assertAlmostEqual(
                    float(pick.comments[0].text.split("=")[-1]), 1.0, 1)

    def test_family_picking_missing_data(self):
        """ Check that this all works okay when one channel has some gaps """
        gappy_data = self.data.copy()
        sr = gappy_data[0].stats.sampling_rate
        mask = np.zeros(len(gappy_data[0].data), dtype=bool)
        gap_start = self.party[0][5].detect_time - 5
        gap_length = 3600
        gap_end = gap_start + gap_length
        gap_start = int(sr * (gap_start - UTCDateTime(0)))
        gap_end = int(sr * (gap_end - UTCDateTime(0)))
        mask[gap_start: gap_end] = np.ones(int(gap_length * sr), dtype=bool)
        gappy_data[0].data = np.ma.masked_array(
            data=gappy_data[0].data, mask=mask)
        catalog_dict = xcorr_pick_family(
            family=self.party[0], stream=gappy_data, shift_len=0.2, plot=False,
            export_cc=False)
        gap = gappy_data.split().get_gaps()
        for event in catalog_dict.values():
            if len(event.picks) != len(gappy_data):
                self.assertEqual(len(event.picks), len(gappy_data) - 1)
                # Check that the event happened to be in the gap
                self.assertTrue(gap[0][4] <= event.picks[0].time <= gap[0][5])
                # Check that there isn't a pick on the channel missing data
                self.assertNotIn('.'.join(gap[0][0:4]),
                                 [pick.waveform_id.get_seed_string()
                                  for pick in event.picks])
            for pick in event.picks:
                self.assertTrue("cc_max=" in pick.comments[0].text)
                self.assertAlmostEqual(
                    float(pick.comments[0].text.split("=")[-1]), 1.0, 1)

    def test_family_picking_with_interpolation(self):
        catalog_dict = xcorr_pick_family(
            family=self.party[0], stream=self.data, shift_len=0.2, plot=False,
            interpolate=True, export_cc=False)
        for event in catalog_dict.values():
            for pick in event.picks:
                self.assertTrue("cc_max=" in pick.comments[0].text)
                self.assertAlmostEqual(
                    float(pick.comments[0].text.split("=")[-1]), 1.0, 1)

    def test_family_picking_with_new_interpolation(self):
        catalog_dict = xcorr_pick_family(
            family=self.party[0], stream=self.data, shift_len=0.2, plot=False,
            interpolate=True, use_new_resamp_method=True, export_cc=False)
        for event in catalog_dict.values():
            for pick in event.picks:
                self.assertTrue("cc_max=" in pick.comments[0].text)
                self.assertAlmostEqual(
                    float(pick.comments[0].text.split("=")[-1]), 1.0, 1)

    def test_lag_calc_api(self):
        detections = [d for f in self.party for d in f]
        templates = [f.template.st for f in self.party]
        template_names = [f.template.name for f in self.party]
        output_cat = lag_calc(
            detections, self.data, template_names, templates,
            shift_len=0.2, min_cc=0.4, min_cc_from_mean_cc_factor=1,
            horizontal_chans=['E', 'N', '1', '2'],
            vertical_chans=['Z'], cores=1, interpolate=False,
            plot=False, export_cc=False)
        self.assertEqual(len(output_cat), len(detections))
        for event in output_cat:
            self.assertEqual(len(event.picks), len(self.data))
            for pick in event.picks:
                self.assertTrue("cc_max=" in pick.comments[0].text)
                self.assertAlmostEqual(
                    float(pick.comments[0].text.split("=")[-1]), 1.0, 1)

    def test_xcorr_pick_family_export_cc(self):
        cc_dir = 'lag_calc_cc_exported'
        xcorr_pick_family(
            family=self.party[0], stream=self.data.copy(), shift_len=0.2,
            plot=False, interpolate=False, export_cc=True, cc_dir=cc_dir)
        cc_files = glob.glob(os.path.join(cc_dir, '*.npy'))
        assert len(cc_files) == len(self.party[0])
        for fcc in cc_files:
            np.load(fcc)
        if os.path.isdir(cc_dir):
            shutil.rmtree(cc_dir)


class SimpleRealDataTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        st = read().detrend().filter("bandpass", freqmin=2, freqmax=20)
        cls.template = st.slice(starttime=st[0].stats.starttime + 4,
                                endtime=st[0].stats.starttime + 10).copy()
        cls.shift_len = 1.0
        cls.detect_streams = []
        for i in range(1, 8):
            cls.detect_streams.append(
                st.slice(starttime=st[0].stats.starttime + i - cls.shift_len,
                         endtime=st[0].stats.starttime + i + 6 + cls.shift_len
                         ).copy())

    def test_correlation_max_and_position(self):
        ccc, chans = _concatenate_and_correlate(
            streams=self.detect_streams, template=self.template, cores=1)
        samp_rate = self.template[0].stats.sampling_rate
        t_start, t_end = (
            self.template[0].stats.starttime, self.template[0].stats.endtime)
        for _ccc, detect_stream in zip(ccc, self.detect_streams):
            d_start, d_end = (detect_stream[0].stats.starttime,
                              detect_stream[0].stats.endtime)
            if d_start <= t_start and d_end >= t_end:
                for ccc_chan in _ccc:
                    self.assertEqual(round(ccc_chan.max(), 5), 1.0)
                    self.assertEqual(ccc_chan.argmax(),
                                     samp_rate * (t_start - d_start))
            else:
                for ccc_chan in _ccc:
                    self.assertNotEqual(round(ccc_chan.max(), 5), 1.0)

    def test_correlation_precision(self):
        """Compare to correlation function outputs"""
        ccc, chans = _concatenate_and_correlate(
            streams=self.detect_streams, template=self.template, cores=1)
        fftw_xcorr_func = get_stream_xcorr("fftw")
        for _ccc, detect_stream in zip(ccc, self.detect_streams):
            fftw_ccc, _, _ = fftw_xcorr_func(
                templates=[self.template], stream=detect_stream, stack=False)
            for chan_ccc, fftw_chan_ccc in zip(_ccc, fftw_ccc[0]):
                self.assertTrue(np.allclose(
                    chan_ccc, fftw_chan_ccc, atol=.00001))


class ShortTests(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        log = logging.getLogger(lag_calc_module.__name__)
        cls._log_handler = MockLoggingHandler(level='DEBUG')
        log.addHandler(cls._log_handler)
        cls.log_messages = cls._log_handler.messages

    def setUp(self):
        self._log_handler.reset()

    def test_error(self):
        with self.assertRaises(LagCalcError):
            raise LagCalcError('Generic error')
        err = LagCalcError('Generic error')
        self.assertEqual('Generic error', err.value)
        self.assertEqual('Generic error', err.__repr__())
        self.assertEqual('LagCalcError: Generic error', err.__str__())

    def test_bad_interp(self):
        ccc = np.array([-0.21483282, -0.59443731, 0.1898917, -0.67516038,
                        0.60129057, -0.71043723,  0.16709118, 0.96839009,
                        1.58283915, -0.3053663])

        _xcorr_interp(ccc, 0.1, use_new_resamp_method=True)
        self.assertEqual(len(self.log_messages['warning']), 1)
        self.assertTrue(
            'not give an accurate result' in self.log_messages['warning'][0])


if __name__ == '__main__':
    unittest.main()



"""

"""

from .archive_access import (  # noqa: F401
    _par_read, _resolved, _badpath, _badlink, _safemembers,
    temporary_directory)
from .matched_filter import (  # noqa: F401
    _tr_spike_test, _spike_test, _total_microsec, _templates_match,
    _test_event_similarity, _remove_duplicates, _moveout, _mad,
    _pickle_stream, _unpickle_stream, extract_from_stream,
    normxcorr2)
from .clients import (get_waveform_client)  # noqa: F401
